#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2020 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""A wrap that notifies about accuracy bug on safe()/load_model() calls."""

from typing import Any
from typing import Dict
from typing import Optional
from typing import TYPE_CHECKING

from thoth.common import get_justification_link as jl

from ...state import State
from ...wrap import Wrap

if TYPE_CHECKING:
    from ..pipeline_builder import PipelineBuilderContext


class TensorFlow23Accuracy(Wrap):
    """A wrap that notifies about accuracy bug on safe()/load_model() calls.

    https://github.com/tensorflow/tensorflow/issues/42045
    https://github.com/keras-team/keras/issues/14181
    https://github.com/tensorflow/tensorflow/commit/5adacc88077ef82f6c4a7f9bb65f9ed89f9d8947
    """

    _JUSTIFICATION = [
        {
            "type": "WARNING",
            "message": "TensorFlow in version 2.3 produces wrong model accuracy when the model is "
            "serialized using `accuracy`, use `sparse_categorical_accuracy` instead",
            "link": jl("tf_42045"),
        }
    ]

    @classmethod
    def should_include(cls, builder_context: "PipelineBuilderContext") -> Optional[Dict[str, Any]]:
        """Include this wrap in adviser."""
        if builder_context.is_included(cls) or not builder_context.is_adviser_pipeline():
            return None

        return {}

    def run(self, state: State) -> None:
        """Notify about accuracy bug in safe()/load_model() calls."""
        tensorflow_any = (
            state.resolved_dependencies.get("tensorflow")
            or state.resolved_dependencies.get("tensorflow-cpu")
            or state.resolved_dependencies.get("tensorflow-gpu")
            or state.resolved_dependencies.get("intel-tensorflow")
        )

        if tensorflow_any is None:
            return None

        if tensorflow_any[1] == "2.3" or tensorflow_any[1].startswith("2.3."):
            state.add_justification(self._JUSTIFICATION)



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2019, 2020 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Penalize stacks with a CVE."""

import logging

from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Set
from typing import Tuple
from typing import TYPE_CHECKING

import attr
from thoth.common import get_justification_link as jl
from thoth.python import PackageVersion
from thoth.storages.exceptions import NotFoundError

from ..exceptions import NotAcceptable
from ..enums import RecommendationType
from ..step import Step
from ..state import State

if TYPE_CHECKING:
    from ..pipeline_builder import PipelineBuilderContext

_LOGGER = logging.getLogger(__name__)


@attr.s(slots=True)
class CvePenalizationStep(Step):
    """Penalization based on CVE being present in stack."""

    CONFIGURATION_DEFAULT = {"cve_penalization": -0.2}
    _JUSTIFICATION_LINK = jl("cve")

    _messages_logged = attr.ib(type=Set[Tuple[str, str, str]], factory=set, init=False)

    def pre_run(self) -> None:
        """Initialize this pipeline unit before running."""
        self._messages_logged.clear()

    @classmethod
    def should_include(cls, builder_context: "PipelineBuilderContext") -> Optional[Dict[str, Any]]:
        """Remove CVEs only for advised stacks."""
        if not builder_context.is_adviser_pipeline():
            return None

        if builder_context.recommendation_type == RecommendationType.LATEST:
            return None

        if not builder_context.is_included(cls):
            return {}

        return None

    def run(self, _: State, package_version: PackageVersion) -> Optional[Tuple[float, List[Dict[str, str]]]]:
        """Penalize stacks with a CVE."""
        try:
            cve_records = self.context.graph.get_python_cve_records_all(
                package_name=package_version.name, package_version=package_version.locked_version,
            )
        except NotFoundError as exc:
            _LOGGER.warning("Package %r in version %r not found: %r", str(exc))
            return None

        if cve_records:
            package_version_tuple = package_version.to_tuple()
            _LOGGER.debug("Found a CVEs for %r: %r", package_version_tuple, cve_records)

            if self.context.recommendation_type == RecommendationType.SECURITY:
                if package_version_tuple not in self._messages_logged:
                    self._messages_logged.add(package_version_tuple)
                    for cve_record in cve_records:
                        _LOGGER.warning(
                            "Skipping including package %r as a CVE %s was found: %s",
                            package_version_tuple,
                            cve_record.get("cve_name") or cve_record.get("cve_id"),
                            cve_record["advisory"],
                        )

                raise NotAcceptable

            penalization = len(cve_records) * self.configuration["cve_penalization"]

            # Note down package causing this CVE.
            for record in cve_records:
                record["package_name"] = package_version.name
                record["link"] = self._JUSTIFICATION_LINK

            return penalization, cve_records

        return None



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2020 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Test removing SciPy from a TensorFlow>=2.1<2.3 stack."""

from itertools import product

import pytest

from thoth.adviser.context import Context
from thoth.adviser.enums import DecisionType
from thoth.adviser.enums import RecommendationType
from thoth.adviser.pipeline_builder import PipelineBuilderContext
from thoth.adviser.state import State
from thoth.adviser.steps import TensorFlowRemoveSciPyStep
from thoth.adviser.exceptions import SkipPackage
from thoth.python import PackageVersion
from thoth.python import Source

from ...base import AdviserTestCase


class TestTensorFlowRemoveSciPyStep(AdviserTestCase):
    """Test related to removing SciPy from a TensorFlow>=2.1<2.3 stack."""

    @pytest.mark.parametrize(
        "recommendation_type", [RecommendationType.STABLE, RecommendationType.PERFORMANCE, RecommendationType.SECURITY]
    )
    def test_include(self, builder_context: PipelineBuilderContext, recommendation_type: RecommendationType) -> None:
        """Test including this pipeline unit."""
        builder_context.decision_type = None
        builder_context.recommendation_type = recommendation_type
        assert "scipy" not in builder_context.project.pipfile.packages
        assert builder_context.is_adviser_pipeline()
        assert TensorFlowRemoveSciPyStep.should_include(builder_context) == {}

    @pytest.mark.parametrize(
        "recommendation_type,decision_type,add_direct_scipy",
        [
            (RecommendationType.LATEST, None, False),
            (RecommendationType.TESTING, None, False),
            (RecommendationType.STABLE, None, True),
            (None, DecisionType.RANDOM, False),
        ],
    )
    def test_no_include(
        self,
        builder_context: PipelineBuilderContext,
        recommendation_type: RecommendationType,
        decision_type: DecisionType,
        add_direct_scipy: bool,
    ) -> None:
        """Test not including this pipeline unit step."""
        builder_context.decision_type = decision_type
        builder_context.recommendation_type = recommendation_type

        if add_direct_scipy:
            builder_context.project.add_package(package_name="scipy", package_version="==1.2.2", develop=False)
            assert "scipy" in builder_context.project.pipfile.packages.packages

        assert TensorFlowRemoveSciPyStep.should_include(builder_context) is None

    @pytest.mark.parametrize(
        "tf_name,tf_version",
        [
            *product(
                ["tensorflow", "tensorflow-cpu", "tensorflow-gpu", "intel-tensorflow"],
                ["2.1.0", "2.1.1", "2.2.0rc0", "2.2.0"],
            )
        ],
    )
    def test_run_skip_package(self, context: Context, state: State, tf_name: str, tf_version: str) -> None:
        """Test removing scipy from a TensorFlow stack."""
        scipy_package_version = PackageVersion(
            name="scipy", version="==1.2.2", develop=False, index=Source("https://pypi.org/simple"),
        )
        tf_package_version = PackageVersion(
            name=tf_name, version=f"=={tf_version}", develop=False, index=Source("https://pypi.org/simple"),
        )

        assert "tensorflow" not in state.resolved_dependencies
        state.resolved_dependencies["tensorflow"] = tf_package_version.to_tuple()
        context.register_package_version(tf_package_version)
        context.dependents["scipy"] = {
            scipy_package_version.to_tuple(): {(tf_package_version.to_tuple(), "rhel", "8", "3.6")}
        }

        with TensorFlowRemoveSciPyStep.assigned_context(context):
            unit = TensorFlowRemoveSciPyStep()
            with pytest.raises(SkipPackage):
                unit.run(state, scipy_package_version)

    @pytest.mark.parametrize(
        "tf_name,tf_version",
        [
            *product(
                ["tensorflow", "tensorflow-cpu", "tensorflow-gpu", "intel-tensorflow"],
                ["2.0.0", "1.9.1", "1.15.3", "2.3.0"],
            )
        ],
    )
    def test_run_noop(self, context: Context, state: State, tf_name: str, tf_version: str) -> None:
        """Test not removing scipy from a TensorFlow stack."""
        scipy_package_version = PackageVersion(
            name="scipy", version="==1.2.2", develop=False, index=Source("https://pypi.org/simple"),
        )
        tf_package_version = PackageVersion(
            name=tf_name, version=f"=={tf_version}", develop=False, index=Source("https://pypi.org/simple"),
        )

        assert "tensorflow" not in state.resolved_dependencies
        state.resolved_dependencies["tensorflow"] = tf_package_version.to_tuple()
        context.register_package_version(tf_package_version)
        context.dependents["scipy"] = {
            scipy_package_version.to_tuple(): {(tf_package_version.to_tuple(), "rhel", "8", "3.6")}
        }

        unit = TensorFlowRemoveSciPyStep()
        with TensorFlowRemoveSciPyStep.assigned_context(context):
            assert unit.run(state, scipy_package_version) is None

    def test_run_deps(self, context: Context, state: State) -> None:
        """Test not removing scipy from a TensorFlow stack if introduced by another dependency."""
        scipy_package_version = PackageVersion(
            name="scipy", version="==1.2.2", develop=False, index=Source("https://pypi.org/simple"),
        )
        tf_package_version = PackageVersion(
            name="tensorflow", version="==2.2.0", develop=False, index=Source("https://pypi.org/simple"),
        )
        another_package_version = PackageVersion(
            name="some-package", version="==1.0.0", develop=False, index=Source("https://pypi.org/simple"),
        )

        assert "tensorflow" not in state.resolved_dependencies
        state.resolved_dependencies["tensorflow"] = tf_package_version.to_tuple()
        state.resolved_dependencies[another_package_version.name] = another_package_version.to_tuple()
        context.register_package_version(tf_package_version)
        context.dependents["scipy"] = {
            scipy_package_version.to_tuple(): {
                (tf_package_version.to_tuple(), "rhel", "8", "3.6"),
                (another_package_version.to_tuple(), "rhel", "8", "3.6"),
            }
        }

        unit = TensorFlowRemoveSciPyStep()
        with TensorFlowRemoveSciPyStep.assigned_context(context):
            assert unit.run(state, scipy_package_version) is None



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2019, 2020 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""A boot to check for version clashes in packages and dev packages."""

import logging
from typing import Optional
from typing import Dict
from typing import Any
from typing import TYPE_CHECKING

import attr

from ..boot import Boot
from ..exceptions import CannotProduceStack

if TYPE_CHECKING:
    from ..pipeline_builder import PipelineBuilderContext

_LOGGER = logging.getLogger(__name__)


@attr.s(slots=True)
class VersionClashBoot(Boot):
    """Check for version clashes in packages and dev-packages.

    Note the implementation does not resolve, so just exact match is checked. Clashes
    during resolution are reported by resolver.
    """

    @classmethod
    def should_include(cls, builder_context: "PipelineBuilderContext") -> Optional[Dict[str, Any]]:
        """Register self, always."""
        if not builder_context.is_included(cls):
            return {}

        return None

    def run(self) -> None:
        """Check for version clash in packages."""
        for package_version in self.context.project.pipfile.packages:
            if not package_version.is_locked():
                continue

            dev_package_version = self.context.project.pipfile.dev_packages.get(package_version.name)

            if dev_package_version is None or not dev_package_version.is_locked():
                continue

            if package_version.locked_version != dev_package_version.locked_version:
                raise CannotProduceStack(
                    f"Package {package_version.name!r} is locked to {package_version.locked_version!r} "
                    f"in packages section, but in dev section it is locked to {package_version.locked_version!r}"
                )



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2020 Kevin Postlethwait
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Test Security Indicator step."""

import flexmock
import pytest

from thoth.adviser.enums import RecommendationType
from thoth.adviser.pipeline_builder import PipelineBuilderContext
from thoth.adviser.steps import SecurityIndicatorStep
from thoth.adviser.exceptions import NotAcceptable
from thoth.python import PackageVersion
from thoth.python import Source
from thoth.storages import GraphDatabase

from ..base import AdviserTestCase


class TestSecurityIndicatorStep(AdviserTestCase):
    """Test different aspects of si pipeline step."""

    _SECURITY_INFO_EXISTS = {
        "severity_high_confidence_high": 1,
        "severity_high_confidence_medium": 5,
        "severity_high_confidence_low": 3,
        "severity_medium_confidence_high": 2,
        "severity_medium_confidence_medium": 0,
        "severity_medium_confidence_low": 8,
        "severity_low_confidence_high": 21,
        "severity_low_confidence_medium": 6,
        "severity_low_confidence_low": 0,
        "number_of_lines_with_code_in_python_files": 692,
    }

    @pytest.mark.parametrize("recommendation_type", [RecommendationType.STABLE, RecommendationType.SECURITY])
    def test_include(self, builder_context: PipelineBuilderContext, recommendation_type: RecommendationType) -> None:
        """Test including this pipeline unit."""
        builder_context.decision_type = None
        builder_context.recommendation_type = recommendation_type
        assert builder_context.is_adviser_pipeline()
        assert SecurityIndicatorStep.should_include(builder_context) == {}

    @pytest.mark.parametrize(
        "recommendation_type", [RecommendationType.LATEST, RecommendationType.PERFORMANCE, RecommendationType.TESTING]
    )
    def test_no_include(self, builder_context: PipelineBuilderContext, recommendation_type,) -> None:
        """Test not including this pipeline unit step."""
        builder_context.decision_type = None
        builder_context.recommendation_type = recommendation_type
        assert SecurityIndicatorStep.should_include(builder_context) is None

    def test_security_indicator_scoring(self) -> None:
        """Make sure we do score security indicators when the info is available."""
        flexmock(GraphDatabase)
        GraphDatabase.should_receive("get_si_aggregated_python_package_version").with_args(
            package_name="flask", package_version="0.12.0", index_url="https://pypi.org/simple"
        ).and_return(self._SECURITY_INFO_EXISTS).once()

        package_version = PackageVersion(
            name="flask", version="==0.12.0", index=Source("https://pypi.org/simple"), develop=False,
        )

        context = flexmock(graph=GraphDatabase())
        with SecurityIndicatorStep.assigned_context(context):
            step = SecurityIndicatorStep()
            result = step.run(None, package_version)

        assert result is not None
        assert isinstance(result, tuple) and len(result) == 2
        assert isinstance(result[0], float)
        assert self.verify_justification_schema(result[1])

    @pytest.mark.parametrize("recommendation_type", [RecommendationType.SECURITY])
    def test_security_indicator_scoring_missing_secure(self, recommendation_type) -> None:
        """Make sure we don't accept package if si info is missing when recommendation is secure."""
        flexmock(GraphDatabase)
        GraphDatabase.should_receive("get_si_aggregated_python_package_version").with_args(
            package_name="flask", package_version="0.12.0", index_url="https://pypi.org/simple"
        ).and_return(None).once()

        package_version = PackageVersion(
            name="flask", version="==0.12.0", index=Source("https://pypi.org/simple"), develop=False,
        )

        context = flexmock(graph=GraphDatabase())
        context.recommendation_type = recommendation_type
        with pytest.raises(NotAcceptable):
            with SecurityIndicatorStep.assigned_context(context):
                step = SecurityIndicatorStep()
                step.run(None, package_version)

    @pytest.mark.parametrize("recommendation_type", [RecommendationType.STABLE])
    def test_security_indicator_scoring_missing_stable(self, recommendation_type) -> None:
        """Make sure package is kept even if no score exists for security indicators and add justification."""
        flexmock(GraphDatabase)
        GraphDatabase.should_receive("get_si_aggregated_python_package_version").with_args(
            package_name="flask", package_version="0.12.0", index_url="https://pypi.org/simple"
        ).and_return(None).once()

        package_version = PackageVersion(
            name="flask", version="==0.12.0", index=Source("https://pypi.org/simple"), develop=False,
        )

        context = flexmock(graph=GraphDatabase())
        context.recommendation_type = recommendation_type
        with SecurityIndicatorStep.assigned_context(context):
            step = SecurityIndicatorStep()
            result = step.run(None, package_version)

        assert result is not None
        assert isinstance(result, tuple) and len(result) == 2
        assert result[0] == 0
        assert len(result[1]) == 1
        assert self.verify_justification_schema(result[1])
        assert result[1][0]["type"] == "WARNING"
        assert (
            result[1][0]["message"] == "flask===0.12.0 on https://pypi.org/simple has no "
            "gathered information regarding security."
        )



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2019, 2020 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Test report as product by a resolver."""

import pytest
import flexmock

from thoth.adviser.pipeline_config import PipelineConfig
from thoth.adviser.product import Product
from thoth.adviser.report import Report
from thoth.common import RuntimeEnvironment

from .base import AdviserTestCase


@pytest.fixture
def pipeline_config() -> PipelineConfig:  # noqa: D401
    """A fixture for a pipeline config."""
    return PipelineConfig(boots=[], sieves=[], steps=[], strides=[], wraps=[])


class TestReport(AdviserTestCase):
    """Test report product by a resolver."""

    def test_stack_info(self, pipeline_config: PipelineConfig) -> None:
        """Test setting a global stack information."""
        report = Report(count=2, pipeline=pipeline_config)
        report.set_stack_info([{"foo": "bar"}])

        assert report.product_count() == 0
        assert report.to_dict() == {
            "pipeline": {"boots": [], "sieves": [], "steps": [], "strides": [], "wraps": [],},
            "products": [],
            "stack_info": [{"foo": "bar"}],
        }

    def test_add_product(self, pipeline_config: PipelineConfig) -> None:
        """Test adding a product to a report."""
        report = Report(count=2, pipeline=pipeline_config)

        product1 = Product(project=None, score=0.42, justification=[], advised_runtime_environment=None)
        report.add_product(product1)
        assert report.product_count() == 1
        assert list(report.iter_products()) == [product1]
        assert list(report.iter_products_sorted()) == [product1]

        product2 = Product(project=None, score=0.0, justification=[], advised_runtime_environment=None)
        report.add_product(product2)
        assert report.product_count() == 2
        assert set(report.iter_products()) == {product1, product2}
        assert list(report.iter_products_sorted()) == [product1, product2]
        assert list(report.iter_products_sorted(reverse=True)) == [product1, product2]
        assert list(report.iter_products_sorted(reverse=False)) == [product2, product1]

        product3 = Product(project=None, score=0.98, justification=[], advised_runtime_environment=None)
        report.add_product(product3)
        assert report.product_count() == 2
        assert set(report.iter_products()) == {product3, product1}
        assert list(report.iter_products_sorted()) == [product3, product1]
        assert list(report.iter_products_sorted(reverse=True)) == [product3, product1]
        assert list(report.iter_products_sorted(reverse=False)) == [product1, product3]

        product4 = Product(project=None, score=0.666, justification=[], advised_runtime_environment=None,)
        report.add_product(product4)
        assert report.product_count() == 2
        assert set(report.iter_products()) == {product3, product4}
        assert list(report.iter_products_sorted()) == [product3, product4]
        assert list(report.iter_products_sorted(reverse=True)) == [product3, product4]
        assert list(report.iter_products_sorted(reverse=False)) == [product4, product3]

    def test_to_dict(self, pipeline_config: PipelineConfig) -> None:
        """Test conversion to a dict."""
        report = Report(count=3, pipeline=pipeline_config)

        project = flexmock()
        project_dict = {"aresto momentum": "avada kedavra"}
        project.should_receive("to_dict").with_args().and_return(project_dict).twice()  # In test and in the report.

        product = Product(
            project=project,
            score=0.666,
            justification=[{"gryffindor": "le gladium leviosa"}],
            advised_runtime_environment=RuntimeEnvironment.from_dict({"python_version": "3.6"}),
        )
        report.add_product(product)

        assert report.product_count() == 1
        assert list(report.iter_products()) == [product]
        assert report.to_dict() == {
            "pipeline": pipeline_config.to_dict(),
            "products": [product.to_dict()],
            "stack_info": None,
        }



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2020 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""A sieve for filtering packages based on constraints.

This sieve will filter out packages in specific versions *if* they occur
in the resolved stack.
"""

import logging
from typing import Any
from typing import Dict
from typing import Optional
from typing import Generator
from typing import TYPE_CHECKING

from packaging.specifiers import Specifier

import attr
from thoth.python import PackageVersion

from ..sieve import Sieve
from ..exceptions import SieveError

if TYPE_CHECKING:
    from ..pipeline_builder import PipelineBuilderContext

_LOGGER = logging.getLogger(__name__)


@attr.s(slots=True)
class VersionConstraintSieve(Sieve):
    """Filter out packages based on version constraints if they occur in the stack."""

    CONFIGURATION_DEFAULT = {"package_name": None, "version_specifier": None}

    _specifier = attr.ib(type=Optional[Specifier], default=None, init=False)

    @classmethod
    def should_include(cls, builder_context: "PipelineBuilderContext") -> Optional[Dict[str, Any]]:
        """Include this sieve only if user explicitly asks for it."""
        return None

    def pre_run(self) -> None:
        """Parse and initialize specifier used."""
        if self.configuration["package_name"] is None:
            raise SieveError(f"Package name for {self.__class__.__name__!r} is not set")

        if self.configuration["version_specifier"] is None:
            raise SieveError(f"Package version specifier for {self.__class__.__name__!r} is not set")

        self._specifier = Specifier(self.configuration["version_specifier"])

    def run(self, package_versions: Generator[PackageVersion, None, None]) -> Generator[PackageVersion, None, None]:
        """Filter out packages based on build time/installation issues.."""
        if self._specifier is None:
            return

        for package_version in package_versions:
            if package_version.name == self.configuration["package_name"] and not self._specifier.contains(
                package_version.locked_version
            ):
                _LOGGER.debug(
                    "Removing package %r based on configured version specifier %r",
                    self.configuration["package_name"],
                    self.configuration["version_specifier"],
                )
                continue

            yield package_version



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2020 Kevin Postlethwait
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Score based on security indicators aggregated."""

from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Set
from typing import Tuple
from typing import TYPE_CHECKING
import logging

import attr
from thoth.common import get_justification_link as jl
from thoth.python import PackageVersion

from ..enums import RecommendationType
from ..exceptions import NotAcceptable
from ..state import State
from ..step import Step


if TYPE_CHECKING:
    from ..pipeline_builder import PipelineBuilderContext


_LOGGER = logging.getLogger(__name__)


@attr.s(slots=True)
class SecurityIndicatorStep(Step):
    """A step that scores a state based on security info aggregated."""

    _logged_packages = attr.ib(type=Set[Tuple[str, str, str]], default=attr.Factory(set), init=False)

    CONFIGURATION_DEFAULT = {
        "high_confidence_weight": 1,
        "medium_confidence_weight": 0.5,
        "low_confidence_weight": 0.1,
        "high_severity_weight": 100,
        "medium_severity_weight": 10,
        "low_severity_weight": 1,
        "si_reward_weight": 0.5,
    }
    _JUSTIFICATION_LINK = jl("security")

    def pre_run(self) -> None:
        """Initialize this pipeline step before running the pipeline."""
        self._logged_packages.clear()

    @classmethod
    def should_include(cls, builder_context: "PipelineBuilderContext") -> Optional[Dict[str, Any]]:
        """Register only if we are explicitly recommending secure stacks."""
        if (
            builder_context.recommendation_type == RecommendationType.SECURITY
            or builder_context.recommendation_type == RecommendationType.STABLE
        ) and not builder_context.is_included(cls):
            return {}

        return None

    @classmethod
    def _generate_justification(cls, name: str, version: str, index: str) -> List[Dict[str, Any]]:
        return [
            {
                "type": "WARNING",
                "message": f"{name}==={version} on {index} has no gathered information regarding security.",
                "link": cls._JUSTIFICATION_LINK,
            }
        ]

    def run(
        self, state: State, package_version: PackageVersion
    ) -> Optional[Tuple[Optional[float], Optional[List[Dict[str, str]]]]]:
        """Score package based on security indicators gathered, do not include if not analyzed."""
        s_info = self.context.graph.get_si_aggregated_python_package_version(
            package_name=package_version.name,
            package_version=package_version.locked_version,
            index_url=package_version.index.url,
        )

        if s_info is None:
            if self.context.recommendation_type == RecommendationType.SECURITY:
                package_version_tuple = package_version.to_tuple_locked()
                if package_version_tuple not in self._logged_packages:
                    self._logged_packages.add(package_version_tuple)
                    _LOGGER.warning(
                        "No security info for %s===%s on %s",
                        package_version.name,
                        package_version.locked_version,
                        package_version.index.url,
                    )
                raise NotAcceptable
            return (
                0,
                self._generate_justification(
                    name=package_version.name, version=package_version.locked_version, index=package_version.index.url
                ),
            )

        s_score = (
            (
                s_info["severity_high_confidence_high"] * self.configuration["high_confidence_weight"]
                + s_info["severity_high_confidence_medium"] * self.configuration["medium_confidence_weight"]
                + s_info["severity_high_confidence_low"] * self.configuration["low_confidence_weight"]
            )
            * self.configuration["high_severity_weight"]
            + (
                s_info["severity_medium_confidence_high"] * self.configuration["high_confidence_weight"]
                + s_info["severity_medium_confidence_medium"] * self.configuration["medium_confidence_weight"]
                + s_info["severity_medium_confidence_low"] * self.configuration["low_confidence_weight"]
            )
            * self.configuration["medium_severity_weight"]
            + (
                s_info["severity_low_confidence_high"] * self.configuration["high_confidence_weight"]
                + s_info["severity_low_confidence_medium"] * self.configuration["medium_confidence_weight"]
                + s_info["severity_low_confidence_low"] * self.configuration["low_confidence_weight"]
            )
            * self.configuration["low_severity_weight"]
        )

        s_score = s_score / s_info["number_of_lines_with_code_in_python_files"]
        if s_score < 1:
            return self.configuration["si_reward_weight"], None
        return self.configuration["si_reward_weight"] / s_score, None



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2020 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Test sieve to filter out Pandas>=1.2 on Python 3.6."""

import flexmock
import pytest

from thoth.adviser.enums import DecisionType
from thoth.adviser.enums import RecommendationType
from thoth.adviser.pipeline_builder import PipelineBuilderContext
from thoth.adviser.sieves import PandasPy36Sieve
from thoth.adviser.context import Context
from thoth.python import PackageVersion
from thoth.python import Source

from ...base import AdviserTestCase


class TestPandasPy36Sieve(AdviserTestCase):
    """Test sieve to filter out Pandas>=1.2 on Python 3.6."""

    @pytest.mark.parametrize(
        "recommendation_type",
        [RecommendationType.STABLE, RecommendationType.PERFORMANCE, RecommendationType.SECURITY,],
    )
    def test_include(self, builder_context: PipelineBuilderContext, recommendation_type: RecommendationType,) -> None:
        """Test including this pipeline unit."""
        builder_context.recommendation_type = recommendation_type
        builder_context.project.runtime_environment.python_version = "3.6"

        assert builder_context.is_adviser_pipeline()
        assert PandasPy36Sieve.should_include(builder_context) == {}

    @pytest.mark.parametrize(
        "recommendation_type,decision_type,python_version",
        [
            (RecommendationType.TESTING, None, "3.6"),
            (RecommendationType.LATEST, None, "3.6"),
            (None, DecisionType.RANDOM, "3.6"),
            (RecommendationType.STABLE, None, "3.9"),
            (RecommendationType.PERFORMANCE, None, "3.8"),
        ],
    )
    def test_no_include(
        self,
        builder_context: PipelineBuilderContext,
        recommendation_type: RecommendationType,
        decision_type: DecisionType,
        python_version: str,
    ) -> None:
        """Test not including this pipeline unit step."""
        builder_context.decision_type = decision_type
        builder_context.recommendation_type = recommendation_type
        builder_context.project.runtime_environment.python_version = python_version
        assert PandasPy36Sieve.should_include(builder_context) is None

    @pytest.mark.parametrize("pandas_version", ["1.2.0", "2.0.0"])
    def test_run(self, context: Context, pandas_version: str) -> None:
        """Test filtering out Pandas that dropped Python 3.6 support."""
        package_version = PackageVersion(
            name="pandas", version=f"=={pandas_version}", develop=False, index=Source("https://pypi.org/simple"),
        )

        unit = PandasPy36Sieve()
        unit.pre_run()
        with PandasPy36Sieve.assigned_context(context):
            assert unit._message_logged is False
            assert list(unit.run(p for p in [package_version])) == []
            assert unit._message_logged is True

    def test_no_filter(self) -> None:
        """Test not filtering packages that can be included."""
        pkg1 = PackageVersion(name="pandas", version="==1.1.2", develop=False, index=Source("https://pypi.org/simple"),)
        pkg2 = PackageVersion(name="pandas", version="==1.0.0", develop=False, index=Source("https://pypi.org/simple"),)
        pkg3 = PackageVersion(
            name="pandas", version="==0.24.2", develop=False, index=Source("https://pypi.org/simple"),
        )

        pkgs = [pkg1, pkg2, pkg3]

        context = flexmock()
        unit = PandasPy36Sieve()
        unit.pre_run()
        with PandasPy36Sieve.assigned_context(context):
            assert unit._message_logged is False
            assert list(unit.run(p for p in pkgs)) == pkgs
            assert unit._message_logged is False



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2020 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""A step that removes SciPy dependency from the dependency listing for TensorFlow>=2.1<=2.3."""

from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import TYPE_CHECKING

import logging

import attr
from thoth.common import get_justification_link as jl
from thoth.python import PackageVersion

from ...enums import RecommendationType
from ...exceptions import SkipPackage
from ...state import State
from ...step import Step

if TYPE_CHECKING:
    from ..pipeline_builder import PipelineBuilderContext

_LOGGER = logging.getLogger(__name__)


@attr.s(slots=True)
class TensorFlowRemoveSciPyStep(Step):
    """A step that removes SciPy dependency from a TensorFlow>2.1<=2.3 stack.

    https://github.com/tensorflow/tensorflow/pull/41866
    https://github.com/tensorflow/tensorflow/issues/40884
    https://github.com/tensorflow/tensorflow/issues/35709
    https://github.com/tensorflow/tensorflow/issues/41941
    https://github.com/tensorflow/tensorflow/pull/40789
    """

    _message_logged = attr.ib(type=bool, default=False, init=False)

    _MESSAGE = "TensorFlow in versions >=2.1<=2.3 stated SciPy as a dependency but it is not used in the codebase"
    _LINK = jl("tf_rm_scipy")

    _RECOMMENDATION_TYPES = {RecommendationType.LATEST, RecommendationType.TESTING}

    @classmethod
    def should_include(cls, builder_context: "PipelineBuilderContext") -> Optional[Dict[Any, Any]]:
        """Include this unit in adviser, except for latest recommendations."""
        if (
            not builder_context.is_adviser_pipeline()
            or builder_context.recommendation_type in cls._RECOMMENDATION_TYPES
        ):
            return None

        if "scipy" in builder_context.project.pipfile.packages.packages:
            # scipy is a direct dependency, it should be always present in the stack.
            return None

        if not builder_context.is_included(cls):
            return {}

        return None

    def pre_run(self) -> None:
        """Initialize this pipeline unit before each run."""
        self._message_logged = False

    def run(
        self, state: State, package_version: PackageVersion
    ) -> Optional[Tuple[Optional[float], Optional[List[Dict[str, str]]]]]:
        """Remove SciPy dependency from a TensorFlow>2.1<=2.3 stack."""
        if package_version.name != "scipy":
            return None

        tensorflow_any = (
            state.resolved_dependencies.get("tensorflow")
            or state.resolved_dependencies.get("tensorflow-cpu")
            or state.resolved_dependencies.get("tensorflow-gpu")
        )

        if not tensorflow_any:
            return None

        tf_package_version = self.context.get_package_version(tensorflow_any)
        if not tf_package_version:
            return None

        tf_release = tf_package_version.semantic_version.release[:2]
        if tf_release < (2, 1) or tf_release >= (2, 3):
            return None

        # Now check what package introduced the SciPy dependency. If it is solely TensorFlow, we can
        # safely remove SciPy from dependencies.
        scipy_dependents = {i[0] for i in self.context.dependents["scipy"][package_version.to_tuple()]}
        introduced_by = scipy_dependents & set(state.resolved_dependencies.values())
        if len(introduced_by) == 1 and next(iter(introduced_by)) == tensorflow_any:
            if not self._message_logged:
                self._message_logged = True
                _LOGGER.warning("%s - see %s", self._MESSAGE, self._LINK)
                self.context.stack_info.append({"type": "WARNING", "message": self._MESSAGE, "link": self._LINK})

            raise SkipPackage

        return None



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2020 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""A sieve to filter out packages based on Python package index configured."""

import logging
from typing import Optional
from typing import Dict
from typing import Any
from typing import Generator
from typing import TYPE_CHECKING

import attr
from thoth.python import PackageVersion
from voluptuous import Required
from voluptuous import Schema

from ..sieve import Sieve

if TYPE_CHECKING:
    from ..pipeline_builder import PipelineBuilderContext

_LOGGER = logging.getLogger(__name__)


@attr.s(slots=True)
class FilterIndexSieve(Sieve):
    """A sieve to filter out packages based on Python package index configured.

    The filtering is done based on configuration supplied - per package-version. The sieve is
    configured using the following pipeline configuration entry:
      {
        "package_name": "tensorflow",
        "index_url": [
          "https://pypi.org/simple",
          "https://tensorflow.pypi.thoth-station.ninja/index/manylinux2010/AVX2/simple/"
        ]
      }
    """

    CONFIGURATION_DEFAULT = {"package_name": None, "index_url": None}
    CONFIGURATION_SCHEMA = Schema({Required("package_name"): str, Required("index_url"): [str],})

    @classmethod
    def should_include(cls, builder_context: "PipelineBuilderContext") -> Optional[Dict[str, Any]]:
        """Register self, never."""
        return None

    def run(self, package_versions: Generator[PackageVersion, None, None]) -> Generator[PackageVersion, None, None]:
        """Filter out packages based on Python package index configured for the given package."""
        for package_version in package_versions:
            if package_version.name != self.configuration["package_name"]:
                yield package_version

            if package_version.index.url not in self.configuration["index_url"]:
                _LOGGER.warning(
                    "Removing package %r in version %r as index used %r does not conform to any supplied %r",
                    package_version.name,
                    package_version.version,
                    package_version.index.url,
                    self.configuration["index_url"],
                )
                continue

            yield package_version



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2020 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""A boot to check Python version configuration used in adviser."""

import logging
from typing import Optional
from typing import Dict
from typing import Any
from typing import TYPE_CHECKING

import attr

from thoth.common import get_justification_link as jl

from ..boot import Boot

if TYPE_CHECKING:
    from ..pipeline_builder import PipelineBuilderContext

_LOGGER = logging.getLogger(__name__)


@attr.s(slots=True)
class PythonVersionBoot(Boot):
    """A boot that checks Python3 configuration used by user."""

    _LINK_PY_VER_PIPFILE = jl("py_version")
    _LINK_PY_VER_THOTH_CONF = jl("py_version")

    @classmethod
    def should_include(cls, builder_context: "PipelineBuilderContext") -> Optional[Dict[str, Any]]:
        """Register self, always for adviser."""
        if not builder_context.is_adviser_pipeline():
            return None

        if not builder_context.is_included(cls):
            return {}

        return None

    def run(self) -> None:
        """Check Python configuration used by user."""
        python_version = self.context.project.runtime_environment.python_version
        pipfile_python_version = self.context.project.pipfile.meta.requires.get("python_version")
        if pipfile_python_version is not None and python_version is None:
            msg = (
                f"No version of Python specified in the configuration, using Python version "
                f"found in Pipfile: {pipfile_python_version!r}"
            )
            _LOGGER.warning("%s - see %s", msg, self._LINK_PY_VER_PIPFILE)
            self.context.project.runtime_environment.python_version = pipfile_python_version
            self.context.stack_info.append({"type": "WARNING", "message": msg, "link": self._LINK_PY_VER_PIPFILE})
        elif python_version is not None and pipfile_python_version is None:
            msg = (
                f"No version of Python specified explicitly, assigning the one found in "
                f"Thoth's configuration: {python_version!r}"
            )
            _LOGGER.warning("%s - see %s", msg, self._LINK_PY_VER_THOTH_CONF)
            self.context.project.pipfile.meta.requires["python_version"] = python_version
            self.context.stack_info.append({"type": "WARNING", "message": msg, "link": self._LINK_PY_VER_THOTH_CONF})



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2019, 2020 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Implementation of pipeline builder - create pipeline configuration based on the project and its library usage."""

import os
import logging
import json
from typing import Any
from typing import Dict
from typing import Generator
from typing import List
from typing import Optional
from typing import Set

import attr
from thoth.python import Project
from thoth.storages import GraphDatabase
import yaml

from .enums import DecisionType
from .enums import RecommendationType
from .exceptions import InternalError
from .exceptions import UnknownPipelineUnitError
from .exceptions import PipelineConfigurationError
from .pipeline_config import PipelineConfig
from .boot import Boot
from .sieve import Sieve
from .step import Step
from .stride import Stride
from .wrap import Wrap
from .unit import Unit

_LOGGER = logging.getLogger(__name__)


@attr.s(slots=True)
class PipelineBuilderContext:
    """A context passed to units to determine if they want to participate in a pipeline."""

    graph = attr.ib(type=GraphDatabase, kw_only=True, default=None)
    project = attr.ib(type=Project, kw_only=True, default=None)
    library_usage = attr.ib(type=Optional[Dict[str, Any]], kw_only=True, default=None)
    decision_type = attr.ib(type=Optional[DecisionType], kw_only=True, default=None)
    recommendation_type = attr.ib(type=Optional[RecommendationType], kw_only=True, default=None)

    _boots = attr.ib(type=List[Boot], default=attr.Factory(list), kw_only=True)
    _sieves = attr.ib(type=List[Sieve], default=attr.Factory(list), kw_only=True)
    _steps = attr.ib(type=List[Step], default=attr.Factory(list), kw_only=True)
    _strides = attr.ib(type=List[Stride], default=attr.Factory(list), kw_only=True)
    _wraps = attr.ib(type=List[Wrap], default=attr.Factory(list), kw_only=True)
    _boots_included = attr.ib(type=Set[Any], default=attr.Factory(set), kw_only=True)
    _sieves_included = attr.ib(type=Set[Any], default=attr.Factory(set), kw_only=True)
    _steps_included = attr.ib(type=Set[Any], default=attr.Factory(set), kw_only=True)
    _strides_included = attr.ib(type=Set[Any], default=attr.Factory(set), kw_only=True)
    _wraps_included = attr.ib(type=Set[Any], default=attr.Factory(set), kw_only=True)

    @property
    def boots(self) -> List[Boot]:
        """Get all boots registered to this pipeline builder context."""
        return self._boots

    @property
    def sieves(self) -> List[Sieve]:
        """Get all sieves registered to this pipeline builder context."""
        return self._sieves

    @property
    def steps(self) -> List[Step]:
        """Get all steps registered to this pipeline builder context."""
        return self._steps

    @property
    def strides(self) -> List[Stride]:
        """Get all strides registered to this pipeline builder context."""
        return self._strides

    @property
    def wraps(self) -> List[Wrap]:
        """Get all wraps registered to this pipeline builder context."""
        return self._wraps

    def __attrs_post_init__(self) -> None:
        """Verify we have only adviser or dependency monkey specific builder."""
        if self.decision_type is not None and self.recommendation_type is not None:
            raise ValueError("Cannot instantiate builder for adviser and dependency monkey at the same time")

        if self.decision_type is None and self.recommendation_type is None:
            raise ValueError("Cannot instantiate builder context not specific to adviser nor dependency monkey")

    def is_included(self, unit_class: type) -> bool:
        """Check if the given pipeline unit is already included in the pipeline configuration."""
        if issubclass(unit_class, Boot):
            return unit_class in self._boots_included
        if issubclass(unit_class, Sieve):
            return unit_class in self._sieves_included
        elif issubclass(unit_class, Step):
            return unit_class in self._steps_included
        elif issubclass(unit_class, Stride):
            return unit_class in self._strides_included
        elif issubclass(unit_class, Wrap):
            return unit_class in self._wraps_included

        raise InternalError(f"Unknown unit {unit_class.__name__!r}")

    def is_adviser_pipeline(self) -> bool:
        """Check if the pipeline built is meant for adviser."""
        return self.decision_type is None and self.recommendation_type is not None

    def is_dependency_monkey_pipeline(self) -> bool:
        """Check if the pipeline built is meant for Dependency Monkey."""
        return self.decision_type is not None and self.recommendation_type is None

    def add_unit(self, unit: Unit) -> None:
        """Add the given unit to pipeline configuration."""
        if isinstance(unit, Boot):
            self._boots_included.add(unit.__class__)
            self._boots.append(unit)
            return
        if isinstance(unit, Sieve):
            self._sieves_included.add(unit.__class__)
            self._sieves.append(unit)
            return
        elif isinstance(unit, Step):
            self._steps_included.add(unit.__class__)
            self._steps.append(unit)
            return
        elif isinstance(unit, Stride):
            self._strides_included.add(unit.__class__)
            self._strides.append(unit)
            return
        elif isinstance(unit, Wrap):
            self._wraps_included.add(unit.__class__)
            self._wraps.append(unit)
            return

        raise InternalError(f"Unknown unit {unit!r} of type {unit.__class__.__name__!r}")


class PipelineBuilder:
    """Builder responsible for creating pipeline configuration from the project and its library usage."""

    __slots__: List[str] = []

    def __init__(self) -> None:
        """Instantiate of the pipeline builder - do NOT instantiate this class."""
        raise NotImplementedError("Cannot instantiate pipeline builder")

    @staticmethod
    def _iter_units() -> Generator[type, None, None]:
        """Iterate over pipeline units available in this implementation."""
        # Imports placed here to simplify tests.
        import thoth.adviser.boots
        import thoth.adviser.sieves
        import thoth.adviser.steps
        import thoth.adviser.strides
        import thoth.adviser.wraps

        for boot_name in thoth.adviser.boots.__all__:
            yield getattr(thoth.adviser.boots, boot_name)

        for sieve_name in thoth.adviser.sieves.__all__:
            yield getattr(thoth.adviser.sieves, sieve_name)

        for step_name in thoth.adviser.steps.__all__:
            yield getattr(thoth.adviser.steps, step_name)

        for stride_name in thoth.adviser.strides.__all__:
            yield getattr(thoth.adviser.strides, stride_name)

        for wrap_name in thoth.adviser.wraps.__all__:
            yield getattr(thoth.adviser.wraps, wrap_name)

    @classmethod
    def _build_configuration(cls, ctx: PipelineBuilderContext) -> PipelineConfig:
        """Instantiate units and return the actual pipeline configuration."""
        # As pipeline steps can have dependencies on each other, iterate over them until we have any change done
        # to the pipeline configuration.
        _LOGGER.info("Creating pipeline configuration")

        blocked_units = (
            set(os.environ["THOTH_ADVISER_BLOCKED_UNITS"].split(","))
            if "THOTH_ADVISER_BLOCKED_UNITS" in os.environ
            else set()
        )

        change = True
        while change:
            change = False
            for unit_class in cls._iter_units():
                if unit_class.__name__ in blocked_units:
                    _LOGGER.debug(
                        "Avoiding adding pipeline unit %r based on blocked units configuration", unit_class.__name__,
                    )
                    continue

                unit_configuration = unit_class.should_include(ctx)  # type: ignore
                if unit_configuration is None:
                    _LOGGER.debug(
                        "Pipeline unit %r will not be included in the pipeline configuration in this round",
                        unit_class.__name__,
                    )
                    continue

                change = True

                _LOGGER.debug(
                    "Including pipeline unit %r in pipeline configuration with unit configuration %r",
                    unit_class.__name__,
                    unit_configuration,
                )
                unit_instance = unit_class()

                if unit_configuration:
                    try:
                        unit_instance.update_configuration(unit_configuration)
                    except Exception as exc:
                        raise PipelineConfigurationError(
                            f"Filed to initialize pipeline unit configuration for {unit_class.__name__!r} "
                            f"with configuration {unit_configuration!r}: {str(exc)}"
                        ) from exc

                ctx.add_unit(unit_instance)

        pipeline = PipelineConfig(
            boots=ctx.boots, sieves=ctx.sieves, steps=ctx.steps, strides=ctx.strides, wraps=ctx.wraps,
        )
        _LOGGER.debug(
            "Pipeline configuration creation ended, configuration:\n%s", json.dumps(pipeline.to_dict(), indent=2),
        )
        return pipeline

    @staticmethod
    def _do_instantiate_from_dict(module: object, configuration_entry: Dict[str, Any]) -> Unit:
        """Instantiate a pipeline unit from a dict representation."""
        if "name" not in configuration_entry:
            raise ValueError(f"No pipeline unit name provided in the configuration entry: {configuration_entry!r}")

        try:
            unit_class = getattr(module, configuration_entry["name"])
        except AttributeError as exc:
            raise UnknownPipelineUnitError(f"Cannot import unit {configuration_entry['name']}: {str(exc)}") from exc

        unit: Unit = unit_class()

        if configuration_entry.get("configuration"):
            try:
                unit.update_configuration(configuration_entry["configuration"])
            except Exception as exc:
                raise PipelineConfigurationError(
                    f"Filed to initialize pipeline unit configuration for {unit_class.__name__!r} "
                    f"with configuration {configuration_entry['configuration']!r}: {str(exc)}"
                ) from exc

        return unit

    @classmethod
    def from_dict(cls, dict_: Dict[str, Any]) -> "PipelineConfig":
        """Instantiate pipeline configuration based on dictionary supplied."""
        # Imports placed here to simplify tests.
        import thoth.adviser.boots
        import thoth.adviser.sieves
        import thoth.adviser.steps
        import thoth.adviser.strides
        import thoth.adviser.wraps

        boots = []
        for boot_entry in dict_.pop("boots", []):
            boots.append(cls._do_instantiate_from_dict(thoth.adviser.boots, boot_entry))

        sieves = []
        for sieve_entry in dict_.pop("sieves", []):
            sieves.append(cls._do_instantiate_from_dict(thoth.adviser.sieves, sieve_entry))

        steps = []
        for step_entry in dict_.pop("steps", []):
            steps.append(cls._do_instantiate_from_dict(thoth.adviser.steps, step_entry))

        strides = []
        for stride_entry in dict_.pop("strides", []):
            strides.append(cls._do_instantiate_from_dict(thoth.adviser.strides, stride_entry))

        wraps = []
        for wrap_entry in dict_.pop("wraps", []):
            wraps.append(cls._do_instantiate_from_dict(thoth.adviser.wraps, wrap_entry))

        if dict_:
            _LOGGER.warning("Unknown entry in pipeline configuration: %r", dict_)

        pipeline = PipelineConfig(
            boots=boots,  # type: ignore
            sieves=sieves,  # type: ignore
            steps=steps,  # type: ignore
            strides=strides,  # type: ignore
            wraps=wraps,  # type: ignore
        )
        _LOGGER.debug(
            "Pipeline configuration creation ended, configuration:\n%s", json.dumps(pipeline.to_dict(), indent=2),
        )
        return pipeline

    @classmethod
    def load(cls, config: str) -> "PipelineConfig":
        """Load pipeline configuration from a file or a string."""
        if os.path.isfile(config):
            _LOGGER.debug("Loading pipeline configuration from file %r", config)
            with open(config, "r") as config_file:
                config = config_file.read()

        return cls.from_dict(yaml.safe_load(config))

    @classmethod
    def get_adviser_pipeline_config(
        cls,
        *,
        recommendation_type: RecommendationType,
        graph: GraphDatabase,
        project: Project,
        library_usage: Optional[Dict[str, Any]],
    ) -> PipelineConfig:
        """Get adviser's pipeline configuration."""
        return cls._build_configuration(
            PipelineBuilderContext(
                graph=graph, project=project, library_usage=library_usage, recommendation_type=recommendation_type,
            )
        )

    @classmethod
    def get_dependency_monkey_pipeline_config(
        cls,
        *,
        decision_type: DecisionType,
        graph: GraphDatabase,
        project: Project,
        library_usage: Optional[Dict[str, Any]],
    ) -> PipelineConfig:
        """Get dependency-monkey's pipeline configuration."""
        return cls._build_configuration(
            PipelineBuilderContext(
                graph=graph, project=project, library_usage=library_usage, decision_type=decision_type,
            )
        )



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2020 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Test TensorFlow AVX2 step."""

from typing import Optional

import flexmock
import pytest

from thoth.adviser.enums import DecisionType
from thoth.adviser.enums import RecommendationType
from thoth.adviser.pipeline_builder import PipelineBuilderContext
from thoth.adviser.steps import TensorFlowAVX2Step
from thoth.python import PackageVersion
from thoth.python import Source

from ...base import AdviserTestCase


class TestTensorFlowAVX2Step(AdviserTestCase):
    """Test TensorFlow AVX2 step recommending AICoE TensorFlow builds."""

    @pytest.mark.parametrize("recommendation_type", [RecommendationType.STABLE, RecommendationType.TESTING])
    def test_include(self, builder_context: PipelineBuilderContext, recommendation_type: RecommendationType) -> None:
        """Test including this pipeline unit."""
        builder_context.decision_type = None
        builder_context.recommendation_type = recommendation_type
        # A Haswell CPU with AVX2 support.
        builder_context.project.runtime_environment.hardware.cpu_family = 0x6
        builder_context.project.runtime_environment.hardware.cpu_model = 0xF
        assert builder_context.is_adviser_pipeline()
        assert TensorFlowAVX2Step.should_include(builder_context) == {}

    @pytest.mark.parametrize(
        "recommendation_type,decision_type,cpu_family,cpu_model",
        [
            (RecommendationType.STABLE, None, None, None),  # No CPU info.
            (RecommendationType.TESTING, None, -1, -1),  # No AVX2 support in CPU.
            (None, DecisionType.RANDOM, 0x6, 0xF),  # A Dependency Monkey run.
        ],
    )
    def test_no_include(
        self,
        builder_context: PipelineBuilderContext,
        recommendation_type,
        decision_type: DecisionType,
        cpu_family: Optional[int],
        cpu_model: Optional[int],
    ) -> None:
        """Test not including this pipeline unit step."""
        builder_context.decision_type = decision_type
        builder_context.recommendation_type = recommendation_type
        # A Haswell CPU with AVX2 support.
        builder_context.project.runtime_environment.hardware.cpu_family = cpu_family
        builder_context.project.runtime_environment.hardware.cpu_model = cpu_model
        assert TensorFlowAVX2Step.should_include(builder_context) is None

    def test_tf_avx2(self) -> None:
        """Test recommending TensorFlow with AVX2 support."""
        package_version = PackageVersion(
            name="tensorflow",
            version="==2.2.0",
            develop=False,
            index=Source("https://tensorflow.pypi.thoth-station.ninja/index/manylinux2010/AVX2/simple"),
        )

        # State and context are unused in the actual pipeline run.
        state, context = flexmock(), flexmock()
        with TensorFlowAVX2Step.assigned_context(context):
            unit = TensorFlowAVX2Step()
            result = unit.run(state, package_version)
            assert len(result[1]) == 1
            assert result[1][0].pop("link", None) is not None, "No link to justification document provided"
            assert result == (
                0.2,
                [
                    {
                        "message": "AICoE TensorFlow builds are optimized for AVX2 instruction "
                        "sets supported in the CPU identified",
                        "type": "INFO",
                    }
                ],
            )
            assert self.verify_justification_schema(result[1])

    def test_no_tf_avx2(self) -> None:
        """Test not recommending TensorFlow without AVX2 support."""
        package_version = PackageVersion(
            name="tensorflow", version="==2.2.0", develop=False, index=Source("https://pypi.org/simple"),
        )

        # State and context are unused in the actual pipeline run.
        state, context = flexmock(), flexmock()
        with TensorFlowAVX2Step.assigned_context(context):
            unit = TensorFlowAVX2Step()
            assert unit.run(state, package_version) is None



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2019 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""A boot to remap UBI to RHEL."""

import logging
from typing import Optional
from typing import Dict
from typing import Any
from typing import TYPE_CHECKING

import attr

from thoth.common import get_justification_link as jl

from ..boot import Boot

if TYPE_CHECKING:
    from ..pipeline_builder import PipelineBuilderContext

_LOGGER = logging.getLogger(__name__)


@attr.s(slots=True)
class UbiBoot(Boot):
    """Remap UBI to RHEL.

    As UBI has ABI compatibility with RHEL, remap any UBI to RHEL.
    """

    _MESSAGE = "Using observations for RHEL instead of UBI, RHEL is ABI compatible with UBI"
    _JUSTIFICATION_LINK = jl("rhel_ubi")

    @classmethod
    def should_include(cls, builder_context: "PipelineBuilderContext") -> Optional[Dict[str, Any]]:
        """Register self, always."""
        if not builder_context.is_included(cls):
            return {}

        return None

    def run(self) -> None:
        """Remap UBI to RHEL as Thoth keeps track of RHEL and UBI is ABI compatible."""
        if self.context.project.runtime_environment.operating_system.name == "ubi":
            _LOGGER.info("%s - see %s", self._MESSAGE, self._JUSTIFICATION_LINK)
            self.context.stack_info.append(
                {"type": "WARNING", "message": self._MESSAGE, "link": self._JUSTIFICATION_LINK}
            )
            self.context.project.runtime_environment.operating_system.name = "rhel"



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2020 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Test version constraint sieve."""

import pytest

from thoth.adviser.sieves import VersionConstraintSieve
from thoth.adviser.exceptions import SieveError
from thoth.python import Source
from thoth.python import PackageVersion

from ..base import AdviserTestCase


class TestVersionConstrainSieve(AdviserTestCase):
    """Test version constraint sieve."""

    def test_configuration_error(self) -> None:
        """Test removing a locked package based on direct dependencies."""
        # Default values should error, no adjustment to values.
        with pytest.raises(SieveError):
            VersionConstraintSieve().pre_run()

        unit = VersionConstraintSieve()
        unit.update_configuration({"package_name": None, "version_specifier": None})
        with pytest.raises(SieveError):
            unit.pre_run()

        unit = VersionConstraintSieve()
        unit.update_configuration(
            {"package_name": "tensorflow", "version_specifier": None,}
        )
        with pytest.raises(SieveError):
            unit.pre_run()

        unit = VersionConstraintSieve()
        unit.update_configuration(
            {"package_name": None, "version_specifier": ">2.0",}
        )
        with pytest.raises(SieveError):
            unit.pre_run()

    def test_default_configuration(self) -> None:
        """Test obtaining default configuration."""
        unit = VersionConstraintSieve()
        assert unit.configuration == {"package_name": None, "version_specifier": None}

    def test_run_filter(self) -> None:
        """Test filtering a package based on version specifier."""
        package_version = PackageVersion(
            name="tensorflow", version="==2.0.0", index=Source("https://pypi.org/simple"), develop=False,
        )
        unit = VersionConstraintSieve()
        unit.update_configuration(
            {"package_name": "tensorflow", "version_specifier": "<2.0",}
        )
        unit.pre_run()
        assert list(unit.run([package_version])) == []

    def test_run_no_filter(self) -> None:
        """Test not filtering a package based on version specifier."""
        package_version = PackageVersion(
            name="tensorboard", version="==2.1.0", index=Source("https://pypi.org/simple"), develop=False,
        )
        unit = VersionConstraintSieve()
        unit.update_configuration(
            {"package_name": "tensorboard", "version_specifier": ">2.0",}
        )
        unit.pre_run()
        assert list(unit.run([package_version])) == [package_version]



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2020 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""A wrap that notifies about adjusting environment variables when Intel's MKL is used.

https://software.intel.com/en-us/mkl-linux-developer-guide-setting-the-number-of-openmp-threads
"""

from typing import TYPE_CHECKING
from typing import Optional, Dict, Any

from thoth.common import get_justification_link as jl

from ...state import State
from ...wrap import Wrap

if TYPE_CHECKING:
    from ..pipeline_builder import PipelineBuilderContext


class MKLThreadsWrap(Wrap):
    """A wrap that notifies about Intel's MKL thread configuration for containerized deployments.

    https://software.intel.com/en-us/mkl-linux-developer-guide-setting-the-number-of-openmp-threads
    """

    _JUSTIFICATION_MKL_ENV = [
        {
            "type": "WARNING",
            "message": "Consider adjusting OMP_NUM_THREADS environment variable for containerized deployments, "
            "one or more libraries use Intel's MKL that does not detect correctly "
            "resource allocation in the cluster",
            "link": jl("mkl_threads"),
        }
    ]

    _JUSTIFICATION_INTEL_TF = [
        {
            "type": "WARNING",
            "message": "Make sure your environment has proper Intel Performance Libraries when using "
            "Intel TensorFlow builds",
            "link": jl("mkl_libs"),
        }
    ]

    # As JSON patch does not have any direct way to test presence of an item in an array, we preprend the configuration.
    _ADVISED_MANIFEST_CHANGE_PREPEND = [
        {
            "apiVersion:": "apps.openshift.io/v1",
            "kind": "DeploymentConfig",
            "patch": {
                "op": "add",
                "path": "/spec/template/spec/containers/0/env/0",
                "value": {"name": "OMP_NUM_THREADS", "value": "1"},
            },
        }
    ]

    @classmethod
    def should_include(cls, builder_context: "PipelineBuilderContext") -> Optional[Dict[Any, Any]]:
        """Include this wrap in adviser, once."""
        if not builder_context.is_included(cls):
            return {}

        return None

    def run(self, state: State) -> None:
        """Check for libraries using PyTorch."""
        if "torch" in state.resolved_dependencies or "pytorch" in state.resolved_dependencies:
            state.add_justification(self._JUSTIFICATION_MKL_ENV)
            state.advised_manifest_changes.append(self._ADVISED_MANIFEST_CHANGE_PREPEND)
        elif "intel-tensorflow" in state.resolved_dependencies:
            state.add_justification(self._JUSTIFICATION_MKL_ENV)
            state.add_justification(self._JUSTIFICATION_INTEL_TF)
            state.advised_manifest_changes.append(self._ADVISED_MANIFEST_CHANGE_PREPEND)



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2019, 2020 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Test scoring (penalization) based on a CVE."""

import flexmock
import pytest

from thoth.adviser.enums import RecommendationType
from thoth.adviser.exceptions import NotAcceptable
from thoth.adviser.steps import CvePenalizationStep
from thoth.python import PackageVersion
from thoth.python import Source
from thoth.storages import GraphDatabase

from ..base import AdviserTestCase


class TestCvePenalizationStep(AdviserTestCase):
    """Test scoring (penalization) based on a CVE."""

    _CASE_CANDIDATES = [
        ("flask", "0.12.0", "https://pypi.org/simple"),
        ("click", "2.0", "https://pypi.org/simple"),
        ("pyyaml", "3.12", "https://pypi.org/simple"),
    ]

    _FLASK_CVE = {
        "advisory": "flask version Before 0.12.3 contains a CWE-20: Improper Input Validation "
        "vulnerability in flask that can result in Large amount of memory usage "
        "possibly leading to denial of service.",
        "cve_name": "CVE-2018-1000656",
        "version_range": "<0.12.3",
        "cve_id": "pyup.io-36388",
    }

    def test_cve_penalization(self) -> None:
        """Make sure a CVE affects stack score."""
        flexmock(GraphDatabase)
        GraphDatabase.should_receive("get_python_cve_records_all").with_args(
            package_name="flask", package_version="0.12.0"
        ).and_return([self._FLASK_CVE]).once()

        package_version = PackageVersion(
            name="flask", version="==0.12.0", index=Source("https://pypi.org/simple"), develop=False,
        )

        context = flexmock(graph=GraphDatabase(), recommendation_type=RecommendationType.TESTING)
        with CvePenalizationStep.assigned_context(context):
            step = CvePenalizationStep()
            result = step.run(None, package_version)

        assert result is not None
        assert isinstance(result, tuple) and len(result) == 2
        assert isinstance(result[0], float)
        assert result[0] == 1 * CvePenalizationStep.CONFIGURATION_DEFAULT["cve_penalization"]
        assert isinstance(result[1], list)
        assert result[1] == [self._FLASK_CVE]
        assert self.verify_justification_schema(result[1])

    def test_no_cve_record(self) -> None:
        """Make sure no CVEs do not affect CVE scoring."""
        flexmock(GraphDatabase)
        GraphDatabase.should_receive("get_python_cve_records_all").with_args(
            package_name="flask", package_version="0.12.0"
        ).and_return([]).once()

        package_version = PackageVersion(
            name="flask", version="==0.12.0", index=Source("https://pypi.org/simple"), develop=False,
        )

        context = flexmock(graph=GraphDatabase())
        with CvePenalizationStep.assigned_context(context):
            step = CvePenalizationStep()
            result = step.run(None, package_version)

        assert result is None

    def test_cve_not_acceptable(self) -> None:
        """Test raising an exception if a secure software stack should be resolved."""
        flexmock(GraphDatabase)
        GraphDatabase.should_receive("get_python_cve_records_all").with_args(
            package_name="flask", package_version="0.12.0"
        ).and_return([self._FLASK_CVE]).once()

        package_version = PackageVersion(
            name="flask", version="==0.12.0", index=Source("https://pypi.org/simple"), develop=False,
        )

        context = flexmock(graph=GraphDatabase(), recommendation_type=RecommendationType.SECURITY)
        step = CvePenalizationStep()
        with CvePenalizationStep.assigned_context(context):
            assert not step._messages_logged
            with pytest.raises(NotAcceptable):
                step.run(None, package_version)

        assert len(step._messages_logged) == 1
        assert ("flask", "0.12.0", "https://pypi.org/simple") in step._messages_logged



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2019, 2020 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Test core utilities provided by pipeline unit base class."""

import pytest

from thoth.adviser.unit import Unit
from thoth.python import PackageVersion
from thoth.python import Source

from .base import AdviserTestCase


class TestUnit(AdviserTestCase):
    """Test unit base - core utilities provided by base class for all pipeline units.."""

    @pytest.mark.parametrize(
        "package_version, expected",
        [
            (
                PackageVersion(
                    name="tensorflow-serving-api",
                    version="1.12.3",
                    index=Source("https://tensorflow.pypi.thoth-station.ninja/index/os/fedora/30/jemalloc/simple"),
                    develop=False,
                ),
                True,
            ),
            (
                PackageVersion(
                    name="tensorflow",
                    version="2.0.0rc2",
                    index=Source("https://tensorflow.pypi.thoth-station.ninja/index/manylinux2010/cuda"),
                    develop=False,
                ),
                True,
            ),
            (
                PackageVersion(
                    name="tensorflow", version="2.0.0rc2", index=Source("https://pypi.org/simple"), develop=False,
                ),
                False,
            ),
        ],
    )
    def test_is_aicoe_release(self, package_version: PackageVersion, expected: bool) -> None:
        """Test checking if the given package is an AICoE package."""
        assert Unit.is_aicoe_release(package_version) is expected

    @pytest.mark.parametrize(
        "package_version, expected",
        [
            (
                PackageVersion(
                    name="tensorflow-serving-api",
                    version="1.12.3",
                    index=Source("https://tensorflow.pypi.thoth-station.ninja/index/os/fedora/30/jemalloc/simple"),
                    develop=False,
                ),
                {"os_name": "fedora", "os_version": "30", "configuration": "jemalloc", "platform_tag": None,},
            ),
            (
                PackageVersion(
                    name="tensorflow",
                    version="1.12.3",
                    index=Source("https://tensorflow.pypi.thoth-station.ninja/index/manylinux2010/cuda/simple"),
                    develop=False,
                ),
                {"os_name": None, "os_version": None, "configuration": "cuda", "platform_tag": "manylinux2010",},
            ),
            (
                PackageVersion(
                    name="tensorflow", version="1.11.0", index=Source("https://pypi.org/simple"), develop=False,
                ),
                None,
            ),
        ],
    )
    def test_get_aicoe_configuration(self, package_version, expected) -> None:
        """Check parsing of AICoE specific configuration encoded in the URL."""
        assert Unit.get_aicoe_configuration(package_version) == expected



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2020 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Suggest not to use TensorFlow 2.2 with tensorflow-probability."""

import attr
from typing import Any
from typing import Optional
from typing import Tuple
from typing import List
from typing import Dict
from typing import TYPE_CHECKING
import logging

from thoth.common import get_justification_link as jl
from thoth.python import PackageVersion

from ...enums import RecommendationType
from ...exceptions import NotAcceptable
from ...state import State
from ...step import Step


if TYPE_CHECKING:
    from ..pipeline_builder import PipelineBuilderContext


_LOGGER = logging.getLogger(__name__)


@attr.s(slots=True)
class TensorFlow22ProbabilityStep(Step):
    """Suggest not to use TensorFlow 2.2 with tensorflow-probability.

    https://github.com/tensorflow/tensorflow/issues/40584
    """

    # Run this step each time, regardless of when tensorflow and tensorflow-probability are resolved.
    MULTI_PACKAGE_RESOLUTIONS = True

    _MESSAGE = "TensorFlow in version 2.2 and tensorflow-probability cause runtime errors"
    _LINK = jl("tf_40584")

    _message_logged = attr.ib(type=bool, default=False, init=False)

    def pre_run(self) -> None:
        """Initialize this pipeline unit before each run."""
        self._message_logged = False

    @classmethod
    def should_include(cls, builder_context: "PipelineBuilderContext") -> Optional[Dict[str, Any]]:
        """Register this pipeline unit for adviser if not using latest recommendations."""
        if (
            not builder_context.is_adviser_pipeline()
            or builder_context.recommendation_type == RecommendationType.LATEST
        ):
            return None

        if not builder_context.is_included(cls):
            return {}

        return None

    def run(
        self, state: State, package_version: PackageVersion
    ) -> Optional[Tuple[Optional[float], Optional[List[Dict[str, str]]]]]:
        """Suggest not to use TensorFlow 2.2 with with tensorflow-probability."""
        if package_version.name != "tensorflow-probability":
            return None

        tensorflow_any = (
            state.resolved_dependencies.get("tensorflow")
            or state.resolved_dependencies.get("tensorflow-cpu")
            or state.resolved_dependencies.get("tensorflow-gpu")
        )

        if not tensorflow_any or (tensorflow_any[1] != "2.2" and not tensorflow_any[1].startswith("2.2.")):
            return None

        if not self._message_logged:
            self._message_logged = True
            _LOGGER.warning("%s - see %s", self._MESSAGE, self._LINK)
            self.context.stack_info.append({"type": "WARNING", "message": self._MESSAGE, "link": self._LINK})

        raise NotAcceptable



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2020 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Test removing backport mock from dependencies."""

from typing import Optional

import pytest

from thoth.adviser.enums import DecisionType
from thoth.adviser.enums import RecommendationType
from thoth.adviser.exceptions import SkipPackage
from thoth.adviser.sieves import MockBackportSieve
from thoth.adviser.context import Context
from thoth.adviser.pipeline_builder import PipelineBuilderContext
from ...base import AdviserTestCase

from thoth.python import PackageVersion
from thoth.python import Source


class TestMockBackportSieve(AdviserTestCase):
    """Test sieve removing mock backport."""

    @pytest.mark.parametrize(
        "python_version,recommendation_type,decision_type,develop",
        [
            ("3.8", RecommendationType.LATEST, None, False),
            ("3.8", None, DecisionType.RANDOM, True),
            ("3.9", RecommendationType.STABLE, None, False),
            ("3.9", RecommendationType.TESTING, None, True),
        ],
    )
    def test_should_include(
        self,
        builder_context: PipelineBuilderContext,
        python_version: str,
        recommendation_type: RecommendationType,
        decision_type: DecisionType,
        develop: bool,
    ) -> None:
        """Test registering this unit."""
        builder_context.project.runtime_environment.python_version = python_version
        builder_context.recommendation_type = recommendation_type
        builder_context.decision_type = decision_type

        assert builder_context.is_dependency_monkey_pipeline() or builder_context.is_adviser_pipeline()
        assert MockBackportSieve.should_include(builder_context) == {}

    @pytest.mark.parametrize(
        "python_version,recommendation_type,decision_type,develop",
        [
            ("3.0", RecommendationType.LATEST, None, False),
            (None, RecommendationType.LATEST, None, False),
            ("3.1", None, DecisionType.RANDOM, True),
            ("3.2", RecommendationType.STABLE, None, False),
            ("3.2", RecommendationType.TESTING, None, True),
            ("2.7", None, DecisionType.ALL, False),
        ],
    )
    def test_no_include(
        self,
        builder_context: PipelineBuilderContext,
        python_version: Optional[str],
        recommendation_type: RecommendationType,
        decision_type: DecisionType,
        develop: bool,
    ) -> None:
        """Test not including this unit."""
        builder_context.project.runtime_environment.python_version = python_version
        builder_context.recommendation_type = recommendation_type
        builder_context.decision_type = decision_type

        assert builder_context.is_dependency_monkey_pipeline() or builder_context.is_adviser_pipeline()
        assert MockBackportSieve.should_include(builder_context) is None

    def test_remove(self, context: Context) -> None:
        """Test removing importlib-resources dependency."""
        package_version = PackageVersion(
            name="mock", version="==4.0.2", develop=False, index=Source("https://pypi.org/simple"),
        )

        unit = MockBackportSieve()
        with MockBackportSieve.assigned_context(context):
            with pytest.raises(SkipPackage):
                list(unit.run(pv for pv in (package_version,)))

            # Run twice to ensure only one entry is added to stack info.
            with pytest.raises(SkipPackage):
                list(unit.run(pv for pv in (package_version,)))

        assert len(context.stack_info) == 1



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2019, 2020 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Test state representation."""

import gc
import pytest
import random
import string
from itertools import chain

from hypothesis import given
from hypothesis.strategies import integers

from thoth.adviser.state import State
from .base import AdviserTestCase


@pytest.fixture
def final_state() -> State:  # noqa: D401
    """A fixture for a final state."""
    state = State(
        score=0.5,
        resolved_dependencies={"daiquiri": ("daiquiri", "1.6.0", "https://pypi.org/simple")},
        unresolved_dependencies={},
        advised_runtime_environment=None,
    )
    state.add_justification(AdviserTestCase.JUSTIFICATION_SAMPLE_1)
    return state


class TestState(AdviserTestCase):
    """Test state representation and manipulation."""

    def test_justification(self, final_state: State) -> None:
        """Test manipulation with state justification."""
        assert final_state.justification == self.JUSTIFICATION_SAMPLE_1
        final_state.add_justification(self.JUSTIFICATION_SAMPLE_2)
        assert final_state.justification == list(chain(self.JUSTIFICATION_SAMPLE_1, self.JUSTIFICATION_SAMPLE_2))

    def test_is_final(self, state: State, final_state: State) -> None:
        """Test checks for final states."""
        assert final_state.is_final()
        assert not state.is_final()
        state.unresolved_dependencies.pop("flask")
        assert state.is_final()

    def test_add_unresolved_dependency(self, final_state: State) -> None:
        """Test adding unresolved dependencies into a state."""
        # This is actually never done in the code (from final state to a non-final state), but
        # let's test turning the switch.
        assert final_state.is_final()
        final_state.add_unresolved_dependency(("selinon", "1.0.0", "https://pypi.org/simple"))
        assert not final_state.is_final()
        final_state.unresolved_dependencies.pop("selinon")
        assert final_state.is_final()

    def test_get_random_first_unresolved_dependency(self) -> None:
        """Test getting random first unresolved dependency."""
        state = State(score=1.0)
        package_tuple1 = ("tensorflow", "2.1.0", "https://pypi.org/simple")
        package_tuple2 = ("selinon", "1.0.0", "https://pypi.org/simple")

        state.add_unresolved_dependency(package_tuple1)

        assert state.get_random_first_unresolved_dependency() is package_tuple1

        state.add_unresolved_dependency(package_tuple2)

        random_state = random.getstate()
        try:
            random.seed(42)
            assert state.get_random_first_unresolved_dependency() is package_tuple1
            assert state.get_random_first_unresolved_dependency() is package_tuple1
            assert state.get_random_first_unresolved_dependency() is package_tuple2
            assert state.get_random_first_unresolved_dependency() is package_tuple1
        finally:
            random.setstate(random_state)

    def test_get_random_first_unresolved_dependency_error(self) -> None:
        """Test raising an error on random first unresolved dependency if no dependency is available."""
        state = State(score=1.0)
        with pytest.raises(IndexError):
            state.get_random_first_unresolved_dependency()

    def test_clone(self, state: State) -> None:
        """Test cloning of states and their memory footprints."""
        cloned_state = state.clone()
        assert cloned_state is not state
        assert cloned_state.score == state.score

        # Check swallow copies.
        assert cloned_state.unresolved_dependencies is not state.unresolved_dependencies
        assert cloned_state.unresolved_dependencies == state.unresolved_dependencies

        for dependency_name in cloned_state.unresolved_dependencies:
            assert (
                cloned_state.unresolved_dependencies[dependency_name] == state.unresolved_dependencies[dependency_name]
            )
            assert (
                cloned_state.unresolved_dependencies[dependency_name]
                is not state.unresolved_dependencies[dependency_name]
            )

        assert cloned_state.resolved_dependencies is not state.resolved_dependencies
        assert cloned_state.resolved_dependencies == state.resolved_dependencies
        assert cloned_state.advised_runtime_environment is not state.advised_runtime_environment
        assert cloned_state.advised_runtime_environment == state.advised_runtime_environment
        assert cloned_state.justification is not state.justification
        assert cloned_state.justification == state.justification
        assert cloned_state.advised_manifest_changes is not state.advised_manifest_changes

    def test_parent(self) -> None:
        """Test referencing parent and weak reference handling."""
        state = State()
        cloned_state = state.clone()

        assert cloned_state.parent is state

        del state
        gc.collect()

        assert cloned_state.parent is None

    @given(integers(min_value=0, max_value=16384),)
    def test_termial_function(self, n: int) -> None:
        """Test termial function."""
        x = State._termial_function(n)
        assert State._termial_function_solution(x) == n

    @given(integers(min_value=0, max_value=4096),)
    def test_termial_function_compute(self, n: int) -> None:
        """Test termial function - compute explicitly results."""
        x = State._termial_function(n)
        assert x == sum(range(n + 1))

    @given(integers(max_value=-1, min_value=-16384),)
    def test_termial_function_error(self, n: int) -> None:
        """Test termial function."""
        with pytest.raises(ValueError):
            State._termial_function(n)

    @given(integers(min_value=1, max_value=16384),)
    def test_termial_function_solution(self, n: int) -> None:
        """Test solution to termial function."""
        x = State._termial_function(n)
        assert State._termial_function_solution(x - 1) <= n <= State._termial_function_solution(x)

    def test_termial_function_solution_zero(self) -> None:
        """Test solution to termial function when x is set to zero."""
        assert State._termial_function_solution(0) == 0

    @given(integers(min_value=1, max_value=16384),)
    def test_random_termial(self, n) -> None:
        """Compute random termial function value."""
        x = State._random_termial(n)
        assert 0 <= x < n

    @given(integers(min_value=-4096, max_value=0),)
    def test_random_termial_error(self, n) -> None:
        """Test error out when random termial is used wih negative or zero value."""
        with pytest.raises(ValueError):
            State._random_termial(n)

    @pytest.mark.parametrize("n", range(5))
    def test_dict_order(self, n: int) -> None:
        """Test relative insertion into dict preserves order.

        This is an implementation detail for Python3.6 dict implementation and
        requirement for Python 3.7. Make sure the running interpreter obeys
        this.
        """
        items = []
        seen = set()
        for i in range(10):
            key = "".join(random.choice(string.ascii_letters) for i in range(n))
            if key in seen:
                continue

            seen.add(key)
            items.append((key, i))

        dict_ = dict(items)
        assert list(dict_.items()) == items, "Order of items in dict is not preserved"

        dict_["foo"] = 3003
        assert list(dict_.items())[-1] == ("foo", 3003), "Last item added to dict is not last obtained"



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2020 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""A sieve to remove backport of importlib-metadata intended for older Python versions."""

import logging
from typing import Any
from typing import Dict
from typing import Generator
from typing import Optional
from typing import TYPE_CHECKING

import attr
from thoth.common import get_justification_link as jl
from thoth.python import PackageVersion

from ...exceptions import SkipPackage
from ...sieve import Sieve

if TYPE_CHECKING:
    from ...pipeline_builder import PipelineBuilderContext

_LOGGER = logging.getLogger(__name__)


@attr.s(slots=True)
class ImportlibMetadataBackportSieve(Sieve):
    """Remove backport of importlib-metadata, available in the standard library starting Python 3.8.

    https://pypi.org/project/importlib-metadata/
    https://docs.python.org/3/library/importlib.metadata.html
    """

    _MESSAGE = (
        "Dependency 'importlib-metadata' removed: importlib.metadata is available in "
        f"Python standard library starting Python 3.8 - see {jl('backports')}"
    )

    _logged = attr.ib(default=False, type=bool, init=False)

    def pre_run(self) -> None:
        """Initialize self before running."""
        self._logged = False

    @classmethod
    def should_include(cls, builder_context: "PipelineBuilderContext") -> Optional[Dict[str, Any]]:
        """Include for Python 3.8 and above for adviser and dependency monkey runs."""
        if builder_context.is_included(cls) or builder_context.project.runtime_environment.python_version is None:
            return None

        if builder_context.project.runtime_environment.get_python_version_tuple() >= (3, 8):
            return {}

        return None

    def run(self, package_versions: Generator[PackageVersion, None, None]) -> Generator[PackageVersion, None, None]:
        """Remove dependency importlib-metadata for newer Python versions."""
        for package_version in package_versions:
            if package_version.name == "importlib-metadata":
                if not self._logged:
                    self.context.stack_info.append({"type": "WARNING", "message": self._MESSAGE})
                    _LOGGER.warning(self._MESSAGE)
                    self._logged = True

                raise SkipPackage

            yield package_version



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2020 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Test wrap that notifies about accuracy bug on safe()/load_model() calls."""

import pytest

from thoth.adviser.state import State
from thoth.adviser.wraps import TensorFlow23Accuracy

from ...base import AdviserTestCase


class TestTensorFlow23Accuracy(AdviserTestCase):
    """Test wrap that notifies about accuracy bug on safe()/load_model() calls."""

    def test_run_noop(self) -> None:
        """Test no justification added if TensorFlow 2.3 is not resolved."""
        state = State()
        assert not state.justification
        assert "tensorflow" not in state.resolved_dependencies

        state.add_resolved_dependency(("tensorflow", "2.2.0", "https://pypi.org/simple"))

        unit = TensorFlow23Accuracy()
        unit.run(state)

        assert len(state.justification) == 0

    @pytest.mark.parametrize("tf_version", ["2.3.0"])
    def test_run(self, tf_version: str) -> None:
        """Test adding justification added if TensorFlow 2.3 is not resolved."""
        state = State()
        assert not state.justification
        assert "tensorflow" not in state.resolved_dependencies

        state.add_resolved_dependency(("tensorflow", tf_version, "https://pypi.org/simple"))

        unit = TensorFlow23Accuracy()
        unit.run(state)

        assert len(state.justification) == 1
        assert set(state.justification[0].keys()) == {"type", "message", "link"}
        assert state.justification[0]["type"] == "WARNING"
        assert state.justification[0]["message"], "No justification message provided"
        assert state.justification[0]["link"], "Empty link to justification document provided"



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2020 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Sieve Pandas>=1.2 on Python 3.6 as Pandas>=1.2 dropped Python 3.6 support."""

import attr
from typing import Any
from typing import Optional
from typing import Generator
from typing import Dict
from typing import TYPE_CHECKING
import logging

from thoth.common import get_justification_link as jl
from thoth.python import PackageVersion

from ...enums import RecommendationType
from ...sieve import Sieve


if TYPE_CHECKING:
    from ..pipeline_builder import PipelineBuilderContext


_LOGGER = logging.getLogger(__name__)


@attr.s(slots=True)
class PandasPy36Sieve(Sieve):
    """A sieve that makes sure Pandas>=1.2 is not used on Python 3.6 or older.

    See https://github.com/pandas-dev/pandas/pull/35214
    """

    _MESSAGE = f"Pandas in versions >=1.2 dropped Python 3.6 support"
    _JUSTIFICATION_LINK = jl("pandas_py36_drop")

    _message_logged = attr.ib(type=bool, default=False, init=False)

    def pre_run(self) -> None:
        """Initialize this pipeline unit before each run."""
        self._message_logged = False

    @classmethod
    def should_include(cls, builder_context: "PipelineBuilderContext") -> Optional[Dict[str, Any]]:
        """Register this pipeline unit for adviser when Python 3.6 is used, not latest/testing recommendations."""
        if not builder_context.is_adviser_pipeline() or builder_context.is_included(cls):
            return None

        if builder_context.recommendation_type in (RecommendationType.LATEST, RecommendationType.TESTING):
            return None

        if builder_context.project.runtime_environment.get_python_version_tuple() > (3, 6):
            # Not a Python 3.6 or older environment.
            return None

        return {}

    def run(self, package_versions: Generator[PackageVersion, None, None]) -> Generator[PackageVersion, None, None]:
        """Do not use Pandas>=1.2 on Python 3.6."""
        for package_version in package_versions:
            if package_version.name != "pandas" or package_version.semantic_version.release[:2] < (1, 2):
                yield package_version
            elif not self._message_logged:
                self._message_logged = True
                _LOGGER.warning("%s - see %s", self._MESSAGE, self._JUSTIFICATION_LINK)
                self.context.stack_info.append(
                    {"type": "WARNING", "message": self._MESSAGE, "link": self._JUSTIFICATION_LINK}
                )



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2019, 2020 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Wrap units implemented in adviser."""

from .no_onservation import NoObservationWrap
from .python import NoSemanticInterpositionWrap
from .tensorflow import TensorFlow23DictSummary
from .tensorflow import TensorFlow23Accuracy
from .tensorflow import IntelTensorFlowWrap
from .tensorflow import MKLThreadsWrap


# Relative ordering of units is relevant, as the order specifies order
# in which the asked to be registered - any dependencies between them
# can be mentioned here.
__all__ = [
    "NoObservationWrap",
    "NoSemanticInterpositionWrap",
    "TensorFlow23DictSummary",
    "TensorFlow23Accuracy",
    "IntelTensorFlowWrap",
    "MKLThreadsWrap",
]



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2020 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""A wrap that notifies about missing observations."""

from typing import TYPE_CHECKING
from typing import Optional, Dict, Any

from thoth.common import get_justification_link as jl

from ..state import State
from ..wrap import Wrap

if TYPE_CHECKING:
    from ..pipeline_builder import PipelineBuilderContext


class NoObservationWrap(Wrap):
    """A wrap that notifies about missing observations."""

    _JUSTIFICATION = [
        {
            "type": "INFO",
            "message": "No observations spotted for this stack in Thoth's database",
            "link": jl("no_observations"),
        }
    ]

    @classmethod
    def should_include(cls, builder_context: "PipelineBuilderContext") -> Optional[Dict[Any, Any]]:
        """Include this wrap in adviser, once."""
        if not builder_context.is_included(cls) and builder_context.is_adviser_pipeline():
            return {}

        return None

    def run(self, state: State) -> None:
        """Check for no observations made on the given state."""
        if not state.justification:
            state.add_justification(self._JUSTIFICATION)



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2019, 2020 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Representation of an advised stack."""

import logging
from typing import Any
from typing import Optional
from typing import Dict
from typing import List
from typing import Tuple
from typing import Set

import attr

from thoth.common import RuntimeEnvironment
from thoth.python import Project
from thoth.python import PackageVersion
from thoth.storages.exceptions import NotFoundError

from .context import Context
from .state import State
from .utils import log_once


_LOGGER = logging.getLogger(__name__)


@attr.s(slots=True, eq=False, order=False)
class Product:
    """A representation of an advised stack."""

    _LOG_HASHES: Set[Tuple[str, str, str]] = set()

    project = attr.ib(type=Project, kw_only=True)
    score = attr.ib(type=float, kw_only=True)
    justification = attr.ib(type=List[Dict[str, str]], kw_only=True)
    advised_runtime_environment = attr.ib(type=Optional[RuntimeEnvironment], default=None, kw_only=True)
    advised_manifest_changes = attr.ib(type=List[List[Dict[str, Any]]], kw_only=True, default=attr.Factory(list))

    @classmethod
    def from_final_state(cls, *, context: Context, state: State) -> "Product":
        """Instantiate advised stack from final state produced by adviser's pipeline."""
        assert state.is_final(), "Instantiating product from a non-final state"

        package_versions_locked = []
        for package_tuple in state.resolved_dependencies.values():
            package_version: PackageVersion = context.get_package_version(package_tuple, graceful=False)

            # Fill package hashes before instantiating the final product.
            if not package_version.hashes:
                # We can re-use already existing package-version - in that case it already keeps hashes from
                # a previous product instantiation.
                hashes = context.graph.get_python_package_hashes_sha256(*package_tuple)
                package_version.hashes = ["sha256:" + h for h in hashes]

                if not package_version.hashes:
                    log_once(_LOGGER, cls._LOG_HASHES, package_tuple, "No hashes found for package %r", package_tuple)

            # Fill environment markers by checking dependencies that introduced this dependency.
            # We do it only if we have no hashes - if hashes are present, the environment marker was
            # already picked (can be set to None if no marker is present).
            # For direct dependencies, dependents can return an empty set (if dependency is not
            # shared with other dependencies) and marker is propagated from PackageVersion registered in
            # Context.register_package_version.
            dependents_tuples = context.dependents[package_tuple[0]][package_tuple]

            # Marker depends based on the stack that was resolved. Do not change package_version directly,
            # rather clone it and used a cloned version not to clash with environment markers.
            environment_markers = []
            for dependent_tuple in dependents_tuples:
                try:
                    marker = context.graph.get_python_environment_marker(
                        *dependent_tuple[0],
                        dependency_name=package_tuple[0],
                        dependency_version=package_tuple[1],
                        os_name=dependent_tuple[1],
                        os_version=dependent_tuple[2],
                        python_version=dependent_tuple[3],
                    )
                except NotFoundError:
                    # This can happen if we do resolution that is agnostic to runtime
                    # environment. In that case a dependency introduced in one runtime
                    # environment does not need to co-exist in another runtime environment considering
                    # marker evaluation result.
                    continue

                if marker and marker not in environment_markers:
                    environment_markers.append(marker)

            if environment_markers:
                if len(environment_markers) > 1:
                    markers = " or ".join(f"({m})" for m in environment_markers)
                else:
                    markers = environment_markers[0]

                package_version = attr.evolve(package_version, markers=markers,)

            package_versions_locked.append(package_version)

        advised_project = Project.from_package_versions(
            packages=list(context.project.iter_dependencies(with_devel=True)),
            packages_locked=package_versions_locked,
            meta=context.project.pipfile.meta,
            runtime_environment=context.project.runtime_environment,
        )

        return cls(
            project=advised_project,
            score=state.score,
            justification=state.justification,
            advised_runtime_environment=state.advised_runtime_environment,
            advised_manifest_changes=state.advised_manifest_changes,
        )

    def to_dict(self) -> Dict[str, Any]:
        """Convert this instance into a dictionary."""
        advised_runtime_environment = None
        if self.advised_runtime_environment:
            advised_runtime_environment = self.advised_runtime_environment.to_dict()

        return {
            "project": self.project.to_dict(),
            "score": self.score,
            "justification": self.justification,
            "advised_runtime_environment": advised_runtime_environment,
            "advised_manifest_changes": self.advised_manifest_changes,
        }



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2020 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Test wrap adding information about Intel's MKL env variable configuration."""

import jsonpatch
import yaml

from thoth.adviser.state import State
from thoth.adviser.wraps import MKLThreadsWrap

from ...base import AdviserTestCase


class TestMKLThreadsWrap(AdviserTestCase):
    """Test Intel's MKL thread env info wrap."""

    _DEPLOYMENT_CONFIG = """\
apiVersion: apps.openshift.io/v1
kind: DeploymentConfig
metadata:
  name: foo
  namespace: some-namespace
spec:
  replicas: 1
  selector:
    service: foo
  template:
    spec:
      containers:
      - image: "foo:latest"
        env:
        - name: APP_MODULE
          value: "foo"
"""

    def test_run_justification_noop(self) -> None:
        """Test no operation when PyTorch is not present."""
        state = State()
        state.add_resolved_dependency(("micropipenv", "0.1.4", "https://pypi.org/simple"))
        assert not state.justification

        unit = MKLThreadsWrap()
        unit.run(state)

        assert len(state.justification) == 0

    def test_run_add_justification(self) -> None:
        """Test adding information Intel's MKL environment variable."""
        state = State()
        state.add_resolved_dependency(("pytorch", "1.4.0", "https://pypi.org/simple"))
        assert len(state.advised_manifest_changes) == 0
        assert not state.justification

        unit = MKLThreadsWrap()
        unit.run(state)

        assert len(state.justification) == 1
        assert set(state.justification[0].keys()) == {"type", "message", "link"}
        assert state.justification[0]["type"] == "WARNING"
        assert state.justification[0]["link"], "Empty link to justification document provided"
        assert len(state.advised_manifest_changes) == 1
        assert state.advised_manifest_changes[0] == [
            {
                "apiVersion:": "apps.openshift.io/v1",
                "kind": "DeploymentConfig",
                "patch": {
                    "op": "add",
                    "path": "/spec/template/spec/containers/0/env/0",
                    "value": {"name": "OMP_NUM_THREADS", "value": "1"},
                },
            }
        ]
        patch = jsonpatch.JsonPatch(obj["patch"] for obj in state.advised_manifest_changes[0])
        deployment_config = yaml.safe_load(self._DEPLOYMENT_CONFIG)
        assert jsonpatch.apply_patch(deployment_config, patch) == {
            "apiVersion": "apps.openshift.io/v1",
            "kind": "DeploymentConfig",
            "metadata": {"name": "foo", "namespace": "some-namespace"},
            "spec": {
                "replicas": 1,
                "selector": {"service": "foo"},
                "template": {
                    "spec": {
                        "containers": [
                            {
                                "env": [
                                    {"name": "OMP_NUM_THREADS", "value": "1"},
                                    {"name": "APP_MODULE", "value": "foo"},
                                ],
                                "image": "foo:latest",
                            }
                        ]
                    }
                },
            },
        }



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2019, 2020 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Pipeline context carried during annealing."""

from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Generator
from typing import Tuple
from typing import Set
import operator
import heapq

import attr

from thoth.python import PackageVersion
from thoth.python import Source
from thoth.python import Project
from thoth.storages import GraphDatabase

from .beam import Beam
from .exceptions import NotFound
from .enums import RecommendationType
from .enums import DecisionType
from .state import State


@attr.s(slots=True)
class Context:
    """Context carried during adviser's pipeline run.

    It's suitable to cache entries such as PackageVersion to optimize memory usage and optimize overhead
    needed - for example for parsing version strings (this is lazily pre-cached in PackageVersion).
    """

    project = attr.ib(type=Project, kw_only=True)
    graph = attr.ib(type=GraphDatabase, kw_only=True)
    library_usage = attr.ib(type=Optional[Dict[str, Any]], kw_only=True)
    limit = attr.ib(type=int, kw_only=True)
    count = attr.ib(type=int, kw_only=True)
    beam = attr.ib(type=Beam, kw_only=True)
    recommendation_type = attr.ib(type=Optional[RecommendationType], kw_only=True, default=None)
    decision_type = attr.ib(type=Optional[DecisionType], kw_only=True, default=None)
    package_versions = attr.ib(
        type=Dict[Tuple[str, str, str], PackageVersion], kw_only=True, default=attr.Factory(dict),
    )
    dependencies = attr.ib(
        type=Dict[str, Dict[Tuple[str, str, str], Set[Tuple[str, str, str]]]], kw_only=True, default=attr.Factory(dict),
    )
    dependents = attr.ib(
        type=Dict[
            str,
            Dict[Tuple[str, str, str], Set[Tuple[Tuple[str, str, str], Optional[str], Optional[str], Optional[str],]],],
        ],
        kw_only=True,
        default=attr.Factory(dict),
    )
    sources = attr.ib(type=Dict[str, Source], kw_only=True, default=attr.Factory(dict))
    iteration = attr.ib(type=int, default=0, kw_only=True)
    cli_parameters = attr.ib(type=Dict[str, Any], kw_only=True, default=attr.Factory(dict))
    stack_info = attr.ib(type=List[Dict[str, Any]], kw_only=True, default=attr.Factory(list))
    accepted_final_states_count = attr.ib(type=int, kw_only=True, default=0)
    discarded_final_states_count = attr.ib(type=int, kw_only=True, default=0)

    _accepted_states = attr.ib(type=List[Tuple[Tuple[float, int], State]], kw_only=True, default=attr.Factory(list),)
    _accepted_states_counter = attr.ib(type=int, kw_only=True, default=0)

    def __attrs_post_init__(self) -> None:
        """Verify we have only adviser or dependency monkey specific context."""
        if self.decision_type is not None and self.recommendation_type is not None:
            raise ValueError("Cannot instantiate context for adviser and dependency monkey at the same time")

        if self.decision_type is None and self.recommendation_type is None:
            raise ValueError("Cannot instantiate context not specific to adviser nor dependency monkey")

    def iter_accepted_final_states(self) -> Generator[State, None, None]:
        """Get accepted final states by resolution pipeline, states are not sorted."""
        return (item[1] for item in self._accepted_states)

    def iter_accepted_final_states_sorted(self, reverse: bool = True) -> Generator[State, None, None]:
        """Get accepted final states by resolution pipeline sorted by score and their precedence."""
        return (item[1] for item in sorted(self._accepted_states, key=operator.itemgetter(0), reverse=reverse))

    def get_package_version(
        self, package_tuple: Tuple[str, str, str], *, graceful: bool = False
    ) -> Optional[PackageVersion]:
        """Get the given package version registered to the context."""
        package_version = self.package_versions.get(package_tuple)
        if package_version is None and not graceful:
            raise NotFound(f"Package {package_tuple!r} not found in the pipeline context")

        return package_version

    def register_package_version(self, package_version: PackageVersion) -> bool:
        """Register the given package version to the context."""
        package_tuple = package_version.to_tuple()
        registered = self.package_versions.get(package_tuple)
        if registered:
            # If the given package is shared in develop and in the main part, make it main stack part.
            registered.develop = registered.develop or package_version.develop
            return True

        # Direct dependency, no dependency introduced this one.
        self._note_dependencies(package_tuple=None, dependency_tuple=package_tuple)
        self.package_versions[package_tuple] = package_version
        return False

    def register_accepted_final_state(self, state: State) -> None:
        """Register an accepted state by the resolution pipeline."""
        # We keep only `count' states as that was requested by pipeline caller.
        item = ((state.score, self._accepted_states_counter), state)
        self._accepted_states_counter -= 1

        if self.count is not None and len(self._accepted_states) >= self.count:
            heapq.heappushpop(self._accepted_states, item)
        else:
            heapq.heappush(self._accepted_states, item)

    def get_top_accepted_final_state(self) -> Optional[State]:
        """Get the best accepted final state so far computed by the resolution pipeline."""
        if not self._accepted_states:
            return None

        result = self._accepted_states[len(self._accepted_states) // 2]
        for item in self._accepted_states[1 + len(self._accepted_states) // 2 :]:
            if result[0] < item[0]:
                result = item

        return result[1]

    def register_package_tuple(
        self,
        package_tuple: Tuple[str, str, str],
        *,
        develop: bool,
        dependent_tuple: Optional[Tuple[str, str, str]] = None,
        extras: Optional[List[str]] = None,
        os_name: Optional[str],
        os_version: Optional[str],
        python_version: Optional[str],
    ) -> PackageVersion:
        """Register the given package tuple to pipeline context and return its package version representative."""
        registered = self.package_versions.get(package_tuple)

        if registered:
            # If the given package is shared in develop and in the main part, make it main stack part.
            registered.develop = registered.develop or develop
            self._note_dependencies(
                dependent_tuple, package_tuple, os_name=os_name, os_version=os_version, python_version=python_version,
            )
            # This method is called solely on transitive dependencies - for those we do not track
            # extras as extras are already resolved by solver runs (pre-computed). Keep extras untouched
            # in this function call.
            return registered

        source = self.sources.get(package_tuple[2])
        if not source:
            source = Source(package_tuple[2])
            self.sources[package_tuple[2]] = source

        package_version = PackageVersion(
            name=package_tuple[0], version="==" + package_tuple[1], index=source, extras=extras, develop=develop,
        )
        self.package_versions[package_tuple] = package_version
        self._note_dependencies(
            dependent_tuple, package_tuple, os_name=os_name, os_version=os_version, python_version=python_version,
        )
        return package_version

    def _note_dependencies(
        self,
        package_tuple: Optional[Tuple[str, str, str]],
        dependency_tuple: Tuple[str, str, str],
        os_name: Optional[str] = None,
        os_version: Optional[str] = None,
        python_version: Optional[str] = None,
    ) -> None:
        """Note down dependencies that were introduced."""
        # Mapping: package tuple -> dependency tuple
        if package_tuple is not None:
            if package_tuple[0] not in self.dependencies:
                self.dependencies[package_tuple[0]] = {}

            if package_tuple not in self.dependencies[package_tuple[0]]:
                self.dependencies[package_tuple[0]][package_tuple] = set()

            self.dependencies[package_tuple[0]][package_tuple].add(dependency_tuple)

        # Reverse mapping: dependency tuple -> package tuple
        if dependency_tuple[0] not in self.dependents:
            self.dependents[dependency_tuple[0]] = {}

        if dependency_tuple not in self.dependents[dependency_tuple[0]]:
            self.dependents[dependency_tuple[0]][dependency_tuple] = set()

        if package_tuple is not None:
            # Note down only if we have some dependency - otherwise dependency
            # tuple is a direct dependency. In that case we don't need to keep
            # track of environments for which the version was resolved.
            self.dependents[dependency_tuple[0]][dependency_tuple].add(
                (package_tuple, os_name, os_version, python_version)
            )

    def is_dependency_monkey(self) -> bool:
        """Check if the current context refers to a dependency monkey run."""
        return self.recommendation_type is None

    def is_adviser(self) -> bool:
        """Check if the current context refers to a dependency monkey run."""
        return self.decision_type is None



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2019, 2020 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""A base class for implementing sieves."""

from itertools import chain
from typing import Any
from typing import Dict
from typing import Generator
from typing import List
from typing import Union

import attr

from .boot import Boot
from .dm_report import DependencyMonkeyReport
from .exceptions import PipelineUnitError
from .sieve import Sieve
from .step import Step
from .stride import Stride
from .wrap import Wrap

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .unit import Unit
    from .report import Report


@attr.s(slots=True)
class PipelineConfig:
    """A configuration of a pipeline for dependency-monkey and for adviser."""

    boots = attr.ib(type=List[Boot], default=attr.Factory(list))
    sieves = attr.ib(type=List[Sieve], default=attr.Factory(list))
    steps = attr.ib(type=List[Step], default=attr.Factory(list))
    strides = attr.ib(type=List[Stride], default=attr.Factory(list))
    wraps = attr.ib(type=List[Wrap], default=attr.Factory(list))

    def to_dict(self) -> Dict[str, List[Dict[str, Any]]]:
        """Return this pipeline configuration in a dict representation."""
        return {
            "boots": [boot.to_dict() for boot in self.boots],
            "sieves": [sieve.to_dict() for sieve in self.sieves],
            "steps": [step.to_dict() for step in self.steps],
            "strides": [stride.to_dict() for stride in self.strides],
            "wraps": [wrap.to_dict() for wrap in self.wraps],
        }

    def iter_units(self):
        # type:('PipelineConfig') -> Generator[Unit, None, None]
        """Iterate over units present in the configuration."""
        yield from chain(self.boots, self.sieves, self.steps, self.strides, self.wraps)

    def iter_units_reversed(self):
        # type:('PipelineConfig') -> Generator[Unit, None, None]
        """Iterate over units present in the configuration in a reversed order."""
        yield from reversed(self.wraps)
        yield from reversed(self.strides)
        yield from reversed(self.steps)
        yield from reversed(self.sieves)
        yield from reversed(self.boots)

    def call_pre_run(self) -> None:
        """Call pre-run method on all units registered in this configuration."""
        for unit in self.iter_units():
            try:
                unit.pre_run()
            except Exception as exc:
                raise PipelineUnitError(
                    "Failed to run pre_run method on unit %r: %s", unit.__class__.__name__, str(exc),
                ) from exc

    def call_post_run(self) -> None:
        """Call post-run method on all units registered in this configuration."""
        for unit in self.iter_units_reversed():
            try:
                unit.post_run()
            except Exception as exc:
                raise PipelineUnitError(
                    "Failed to run post_run method on unit %r: %s", unit.__class__.__name__, str(exc),
                ) from exc

    def call_post_run_report(self, report):
        # type:('PipelineConfig', Union[Report, DependencyMonkeyReport]) -> None
        """Call post-run method when report is generated."""
        for unit in self.iter_units_reversed():
            try:
                unit.post_run_report(report)
            except Exception as exc:
                raise PipelineUnitError(
                    "Failed to run pre_run_report method on unit %r: %s", unit.__class__.__name__, str(exc),
                ) from exc



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2020 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Test TensorFlow 2.1 urllib3 step."""

import flexmock
import pytest

from thoth.adviser.context import Context
from thoth.adviser.exceptions import NotAcceptable
from thoth.adviser.steps import TensorFlow21Urllib3Step
from thoth.adviser.state import State
from thoth.python import PackageVersion
from thoth.python import Source

from ...base import AdviserTestCase


class TestTensorFlow21Urllib32Step(AdviserTestCase):
    """Test a step that suggests not to use TensorFlow 2.1 as issues with six were spotted on imports."""

    @pytest.mark.parametrize(
        "urllib3_version,tf_version",
        [("1.2", "2.1"), ("1.2.1", "2.1.1"), ("1.2.2", "2.1.0"), ("1.3", "2.1"), ("1.4", "2.1"), ("1.5", "2.1.dev")],
    )
    def test_tf_21(self, context: Context, urllib3_version: str, tf_version: str) -> None:
        """Test penalizing TensorFlow in version 2.1."""
        tf_package_version = PackageVersion(
            name="tensorflow", version=f"=={tf_version}", develop=False, index=Source("https://pypi.org/simple"),
        )

        urllib3_package_version = PackageVersion(
            name="urllib3", version=f"=={urllib3_version}", develop=False, index=Source("https://pypi.org/simple"),
        )

        state = State()
        state.resolved_dependencies["tensorflow"] = tf_package_version.to_tuple()

        assert not context.stack_info

        with TensorFlow21Urllib3Step.assigned_context(context):
            unit = TensorFlow21Urllib3Step()
            unit.pre_run()
            assert unit._message_logged is False
            with pytest.raises(NotAcceptable):
                assert unit.run(state, urllib3_package_version)
                assert unit._message_logged is True

        assert context.stack_info
        assert self.verify_justification_schema(context.stack_info)

    @pytest.mark.parametrize("urllib3_version,tf_version", [("1.2", "2.2.0"), ("1.25.10", "2.1")])
    def test_no_tf_21(self, urllib3_version: str, tf_version: str) -> None:
        """Test no penalization for TensorFlow other than 2.1 by this pipeline step."""
        urllib3_package_version = PackageVersion(
            name="urllib3", version=f"=={urllib3_version}", develop=False, index=Source("https://pypi.org/simple"),
        )

        tf_package_version = PackageVersion(
            name="tensorflow", version=f"=={tf_version}", develop=False, index=Source("https://pypi.org/simple"),
        )

        state = State()
        state.resolved_dependencies["tensorflow"] = tf_package_version.to_tuple()

        # Context is not used during the actual pipeline run.
        context = flexmock()
        with TensorFlow21Urllib3Step.assigned_context(context):
            unit = TensorFlow21Urllib3Step()
            assert unit.run(state, urllib3_package_version) is None



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2020 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Test implementation of Temporal Difference (TD) based predictor with annealing schedule."""

import math
import random

from hypothesis import given
from hypothesis.strategies import floats
from hypothesis.strategies import integers
import flexmock
import pytest

from thoth.adviser.context import Context
from thoth.adviser.predictors import TemporalDifference
from thoth.adviser.predictors import AdaptiveSimulatedAnnealing
from thoth.adviser.state import State
import thoth.adviser.predictors.td as td_module

from ..base import AdviserTestCase


class TestTemporalDifference(AdviserTestCase):
    """Test implementation of Temporal Difference (TD) based predictor with annealing schedule."""

    @given(
        floats(allow_nan=False, allow_infinity=False),
        floats(allow_nan=False, allow_infinity=False),
        floats(min_value=0.0, allow_nan=False, allow_infinity=False),
    )
    def test_acceptance_probability(self, top_score: float, neighbour_score: float, temperature: float) -> None:
        """Test acceptance probability is always between 0 and 1."""
        acceptance_probability = TemporalDifference._compute_acceptance_probability(
            top_score=top_score, neighbour_score=neighbour_score, temperature=temperature,
        )
        assert 0.0 <= acceptance_probability <= 1.0, "Acceptance probability not within 0 and 1"

    @given(
        floats(min_value=0.0, allow_nan=False, allow_infinity=False),
        integers(min_value=0),
        integers(min_value=1),
        integers(min_value=0),
        integers(min_value=0),
    )
    def test_temperature_function(
        self, t0: float, accepted_final_states_count: int, limit: int, iteration: int, count: int,
    ) -> None:
        """Test the temperature function never drops bellow 0."""
        context = flexmock(
            accepted_final_states_count=accepted_final_states_count,
            limit=limit,
            iteration=iteration,
            count=count,
            beam=flexmock(size=96),
        )

        predictor = TemporalDifference()
        assert predictor._temperature_function(t0=t0, context=context) >= 0.0, "Temperature dropped bellow 0 or is NaN"

    def test_init(self) -> None:
        """Test instantiation."""
        predictor = TemporalDifference()
        assert predictor._policy == {}
        assert predictor._temperature_history == []
        assert predictor._temperature == 0.0

    def test_pre_run(self) -> None:
        """Test initialization done before running."""
        predictor = TemporalDifference()

        predictor._policy = {("tensorflow", "2.0.0", "https://pypi.org/simple"): [1.0, 2]}
        predictor._temperature_history = [(0.212, True, 0.23, 100)]
        predictor._temperature = 12.3

        context = flexmock(limit=42)
        with predictor.assigned_context(context):
            predictor.pre_run()

        assert predictor._policy == {}
        assert predictor._temperature_history == []
        assert isinstance(predictor._temperature, float)
        assert predictor._temperature == float(context.limit)

    @pytest.mark.parametrize("float_case", [math.nan, math.inf, -math.inf])
    def test_set_reward_signal_nan_inf(self, float_case: float) -> None:
        """Test (not) keeping the reward signal for nan/inf."""
        predictor = TemporalDifference()
        state = flexmock()
        assert (
            predictor.set_reward_signal(state, ("tensorflow", "2.0.0", "https://pypi.org/simple"), float_case) is None
        )
        assert predictor._policy == {}

    def test_set_reward_signal_unseen(self) -> None:
        """Test keeping the reward signal for an unseen step."""
        reward = 42.24
        package_tuple = ("tensorflow", "2.0.0", "https://thoth-station.ninja")

        state = flexmock()
        state.should_receive("iter_resolved_dependencies").and_return([package_tuple]).once()

        predictor = TemporalDifference()
        predictor._policy = {
            ("numpy", "1.0.0", "https://pypi.org/simple"): [30.30, 92],
        }

        predictor.set_reward_signal(state, None, reward)

        assert predictor._policy == {
            package_tuple: [42.24, 1],
            ("numpy", "1.0.0", "https://pypi.org/simple"): [30.30, 92],
        }

    def test_set_reward_signal_seen(self) -> None:
        """Test keeping the reward signal for a seen step."""
        reward = 24.42
        package_tuple = ("pytorch", "1.0.2", "https://thoth-station.ninja")

        state = flexmock()
        state.should_receive("iter_resolved_dependencies").and_return([package_tuple]).once()

        predictor = TemporalDifference()
        predictor._policy = {
            package_tuple: [16.23, 2010],
            ("numpy", "1.0.0", "https://pypi.org/simple"): [30.30, 92],
        }

        predictor.set_reward_signal(state, None, reward)

        assert predictor._policy == {
            package_tuple: [16.23 + reward, 2011],
            ("numpy", "1.0.0", "https://pypi.org/simple"): [30.30, 92],
        }

    def test_do_exploitation(self) -> None:
        """Tests on exploitation computation."""
        predictor = TemporalDifference()
        predictor._policy = {
            ("tensorflow", "2.1.0", "https://thoth-station.ninja"): [2020.21, 666],
            ("tensorflow", "2.0.0", "https://thoth-station.ninja"): [16.61, 1992],
            ("numpy", "1.0.0", "https://pypi.org/simple"): [30.30, 92],
        }

        state = flexmock()
        state.should_receive("iter_unresolved_dependencies").and_return(
            [
                ("spacy", "2.2.4", "https://pypi.org/simple"),
                ("numpy", "1.0.0", "https://pypi.org/simple"),
                ("tensorflow", "2.1.0", "https://thoth-station.ninja"),
            ]
        ).once()

        state.should_receive("get_random_unresolved_dependency").times(0)
        assert predictor._do_exploitation(state) == ("tensorflow", "2.1.0", "https://thoth-station.ninja",)

    def test_do_exploitation_no_records(self) -> None:
        """Tests on exploitation when no relevant records found."""
        predictor = TemporalDifference()
        assert predictor._policy == {}

        random_unresolved_dependency = (("tensorflow", "2.1.0", "https://thoth-station.ninja"),)

        state = flexmock()
        state.should_receive("iter_unresolved_dependencies").and_return(
            [("micropipenv", "0.1.4", "https://pypi.org/simple"), random_unresolved_dependency,]
        ).once()

        state.should_receive("get_random_unresolved_dependency").with_args(prefer_recent=True).and_return(
            random_unresolved_dependency
        ).once()

        assert predictor._do_exploitation(state) == random_unresolved_dependency

    def test_policy_size_no_shrink(self, context: Context) -> None:
        """Test limiting policy size over runs."""
        predictor = TemporalDifference()

        predictor._policy = {
            ("numpy", "2.0.0", "https://pypi.org/simple"): [1.0, 100],
            ("tensorflow", "2.0.0", "https://thoth-station.ninja/simple"): [2.0, 100],
        }

        rewarded = list(predictor._policy.keys())[0]  # numpy
        state = flexmock()
        state.should_receive("iter_resolved_dependencies").with_args().and_return([rewarded]).once()

        # No shrink as we are in this iteration.
        context.iteration = td_module._TD_POLICY_SIZE_CHECK_ITERATION + 1
        old_policy_size = td_module._TD_POLICY_SIZE
        with predictor.assigned_context(context):
            try:
                td_module._TD_POLICY_SIZE = 1
                predictor.set_reward_signal(state, rewarded, 1.0)
            finally:
                td_module._TD_POLICY_SIZE = old_policy_size

        assert predictor._policy == {
            ("numpy", "2.0.0", "https://pypi.org/simple"): [1.0 + 1.0, 100 + 1],
            ("tensorflow", "2.0.0", "https://thoth-station.ninja/simple"): [2.0, 100],
        }

    def test_policy_size_shrink(self, context: Context) -> None:
        """Test limiting policy size over runs."""
        predictor = TemporalDifference()

        predictor._policy = {
            ("numpy", "2.0.0", "https://pypi.org/simple"): [1.0, 100],
            ("tensorflow", "2.0.0", "https://thoth-station.ninja/simple"): [3.0, 100],
        }

        rewarded = list(predictor._policy.keys())[0]  # numpy
        state = flexmock()
        state.should_receive("iter_resolved_dependencies").with_args().and_return([rewarded]).once()

        # No shrink as we are in this iteration.
        context.iteration = 2 * td_module._TD_POLICY_SIZE_CHECK_ITERATION
        old_policy_size = td_module._TD_POLICY_SIZE
        with predictor.assigned_context(context):
            try:
                td_module._TD_POLICY_SIZE = 1
                predictor.set_reward_signal(state, rewarded, 0.5)
            finally:
                td_module._TD_POLICY_SIZE = old_policy_size

        # the numpy entry with a value of [1.5, 101] gets removed
        assert predictor._policy == {
            ("tensorflow", "2.0.0", "https://thoth-station.ninja/simple"): [3.0, 100],
        }

    def test_run_exploration(self, context: Context) -> None:
        """Tests run when exploration is performed."""
        flexmock(TemporalDifference)
        flexmock(TemporalDifference)
        flexmock(AdaptiveSimulatedAnnealing)

        flexmock(State)
        max_state = State(score=3.0)
        probable_state = State(score=2.0)

        context.beam.add_state(max_state)
        context.beam.add_state(probable_state)

        unresolved_dependency = (
            "pytorch",
            "1.0.0",
            "https://thoth-station.ninja/simple",
        )

        flexmock(random)
        random.should_receive("randrange").with_args(1, 2).and_return(0).once()
        random.should_receive("random").with_args().and_return(
            0.50
        ).once()  # *lower* than acceptance_probability that is 0.75 so we do exploitation
        probable_state.should_receive("get_random_unresolved_dependency").with_args(prefer_recent=True).and_return(
            unresolved_dependency
        ).once()
        TemporalDifference.should_receive("_temperature_function").with_args(1.0, context).and_return(0.9).once()
        AdaptiveSimulatedAnnealing.should_receive("_compute_acceptance_probability").with_args(
            max_state.score, probable_state.score, 0.9
        ).and_return(0.75).once()
        context.beam.should_receive("max").with_args().and_return(max_state).and_return(max_state).twice()

        predictor = TemporalDifference()
        predictor._temperature = 1.0
        with predictor.assigned_context(context):
            assert predictor.run() == (probable_state, unresolved_dependency)

    def test_run_exploitation(self, context: Context) -> None:
        """Tests run when exploitation is performed."""
        flexmock(TemporalDifference)
        flexmock(AdaptiveSimulatedAnnealing)

        max_state = State(score=3.0)
        probable_state = State(score=2.0)

        context.beam.add_state(max_state)
        context.beam.add_state(probable_state)

        unresolved_dependency = (
            "pytorch",
            "1.0.0",
            "https://thoth-station.ninja/simple",
        )

        flexmock(random)
        random.should_receive("randrange").with_args(1, 2).and_return(0).once()
        random.should_receive("random").with_args().and_return(
            0.99
        ).once()  # *higher* than acceptance_probability that is 0.75 so we do exploitation
        TemporalDifference.should_receive("_do_exploitation").with_args(max_state).and_return(
            unresolved_dependency
        ).once()
        TemporalDifference.should_receive("_temperature_function").with_args(1.0, context).and_return(0.9).once()
        AdaptiveSimulatedAnnealing.should_receive("_compute_acceptance_probability").with_args(
            max_state.score, probable_state.score, 0.9
        ).and_return(0.75).once()
        context.beam.should_receive("max").with_args().and_return(max_state).and_return(max_state).twice()

        predictor = TemporalDifference()
        predictor._temperature = 1.0
        with predictor.assigned_context(context):
            assert predictor.run() == (max_state, unresolved_dependency)
